#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2015 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
import ast
import errno
import os
import time
import traceback
import sys
import threading
from typing import Dict, Optional, Tuple, Callable, Union, Sequence, Mapping, TYPE_CHECKING
from base64 import b64decode, b64encode
import json
import socket
import stat

import aiohttp
from aiohttp import web, client_exceptions
from aiorpcx import ignore_after

from . import util
from .network import Network
from .util import (
    json_decode, to_bytes, to_string, profiler, standardize_path, constant_time_compare, InvalidPassword,
    log_exceptions, randrange, OldTaskGroup, UserFacingException, JsonRPCError, os_chmod
)
from .wallet import Wallet, Abstract_Wallet
from .storage import WalletStorage
from .wallet_db import WalletDB, WalletUnfinished
from .commands import known_commands, Commands
from .simple_config import SimpleConfig
from .exchange_rate import FxThread
from .logging import get_logger, Logger
from . import GuiImportError
from .plugin import run_hook, Plugins

if TYPE_CHECKING:
    from electrum import gui


_logger = get_logger(__name__)


class DaemonNotRunning(Exception):
    pass


def get_rpcsock_defaultpath(config: SimpleConfig):
    return os.path.join(config.path, 'daemon_rpc_socket')


def get_rpcsock_default_type(config: SimpleConfig):
    if config.RPC_PORT:
        return 'tcp'
    # Use unix domain sockets when available,
    # with the extra paranoia that in case windows "implements" them,
    # we want to test it before making it the default there.
    if hasattr(socket, 'AF_UNIX') and sys.platform != 'win32':
        return 'unix'
    return 'tcp'


def get_lockfile(config: SimpleConfig):
    return os.path.join(config.path, 'daemon')


def remove_lockfile(lockfile):
    os.unlink(lockfile)


def get_file_descriptor(config: SimpleConfig):
    '''Tries to create the lockfile, using O_EXCL to
    prevent races.  If it succeeds, it returns the FD.
    Otherwise, try and connect to the server specified in the lockfile.
    If this succeeds, the server is returned.  Otherwise, remove the
    lockfile and try again.'''
    lockfile = get_lockfile(config)
    while True:
        try:
            return os.open(lockfile, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o644)
        except OSError:
            pass
        try:
            request(config, 'ping')
            return None
        except DaemonNotRunning:
            # Couldn't connect; remove lockfile and try again.
            remove_lockfile(lockfile)


def request(config: SimpleConfig, endpoint, args=(), timeout: Union[float, int] = 60):
    lockfile = get_lockfile(config)
    for attempt in range(5):
        create_time = None  # type: Optional[float | int]
        path = None
        try:
            with open(lockfile) as f:
                socktype, address, create_time = ast.literal_eval(f.read())
                int(create_time)  # raise if not numeric
                if socktype == 'unix':
                    path = address
                    (host, port) = "127.0.0.1", 0
                    # We still need a host and port for e.g. HTTP Host header
                elif socktype == 'tcp':
                    (host, port) = address
                else:
                    raise Exception(f"corrupt lockfile; socktype={socktype!r}")
        except Exception:
            raise DaemonNotRunning()
        rpc_user, rpc_password = get_rpc_credentials(config)
        server_url = 'http://%s:%d' % (host, port)
        auth = aiohttp.BasicAuth(login=rpc_user, password=rpc_password)
        loop = util.get_asyncio_loop()

        async def request_coroutine(
            *, socktype=socktype, path=path, auth=auth, server_url=server_url, endpoint=endpoint,
        ):
            if socktype == 'unix':
                connector = aiohttp.UnixConnector(path=path)
            elif socktype == 'tcp':
                connector = None # This will transform into TCP.
            else:
                raise Exception(f"impossible socktype ({socktype!r})")
            async with aiohttp.ClientSession(auth=auth, connector=connector) as session:
                c = util.JsonRPCClient(session, server_url)
                return await c.request(endpoint, *args)

        try:
            fut = asyncio.run_coroutine_threadsafe(request_coroutine(), loop)
            return fut.result(timeout=timeout)
        except aiohttp.client_exceptions.ClientConnectorError as e:
            _logger.info(f"failed to connect to JSON-RPC server {e}")
            # We cannot communicate with the daemon.
            # If daemon's creation time is very recent, it might still be starting up.
            # In any other case, we raise: - too old create_time means daemon is likely dead,
            #                              - create_time in future means our clock cannot be trusted.
            if not (create_time <= time.time() <= create_time + 1.0):
                raise DaemonNotRunning()
        # Sleep a bit and try again; daemon might have just been started
        time.sleep(1.0)
    # how did we even get here?! the clock must be going haywire.
    _logger.error(f"Failed to connect to JSON-RPC server. Exhausted all attempts.")
    raise DaemonNotRunning()


def wait_until_daemon_becomes_ready(*, config: SimpleConfig, timeout=5) -> bool:
    t0 = time.monotonic()
    while True:
        if time.monotonic() > t0 + timeout:
            return False  # timeout
        try:
            request(config, 'ping')
            return True  # success
        except DaemonNotRunning:
            time.sleep(0.05)
            continue


def get_rpc_credentials(config: SimpleConfig) -> Tuple[str, str]:
    rpc_user = config.RPC_USERNAME or None
    rpc_password = config.RPC_PASSWORD or None
    # note: we explicitly forbid empty/unset password, and will generate one now instead
    if rpc_user is None or rpc_password is None:
        rpc_user = 'user'
        bits = 128
        nbytes = bits // 8 + (bits % 8 > 0)
        pw_int = randrange(pow(2, bits))
        pw_b64 = b64encode(
            pw_int.to_bytes(nbytes, 'big'), b'-_')
        rpc_password = to_string(pw_b64, 'ascii')
        config.RPC_USERNAME = rpc_user
        config.RPC_PASSWORD = rpc_password
    return rpc_user, rpc_password


class AuthenticationError(Exception):
    pass


class AuthenticationInvalidOrMissing(AuthenticationError):
    pass


class AuthenticationCredentialsInvalid(AuthenticationError):
    pass


class AuthenticatedServer(Logger):

    def __init__(self, rpc_user, rpc_password):
        Logger.__init__(self)
        self.rpc_user = rpc_user
        self.rpc_password = rpc_password
        self.auth_lock = asyncio.Lock()
        self._methods = {}  # type: Dict[str, Callable]

    def register_method(self, name: str, f):
        assert name not in self._methods, f"name collision for {name}"
        self._methods[name] = f

    async def authenticate(self, headers):
        if not self.rpc_password:
            raise Exception('Server RPC password is unset. This should not happen.')
        auth_string = headers.get('Authorization', None)
        if auth_string is None:
            raise AuthenticationInvalidOrMissing('CredentialsMissing')
        basic, _, encoded = auth_string.partition(' ')
        if basic != 'Basic':
            raise AuthenticationInvalidOrMissing('UnsupportedType')
        encoded = to_bytes(encoded, 'utf8')
        credentials = to_string(b64decode(encoded, validate=True), 'utf8')
        username, _, password = credentials.partition(':')
        if not (constant_time_compare(username, self.rpc_user)
                and constant_time_compare(password, self.rpc_password)):
            await asyncio.sleep(0.050)
            raise AuthenticationCredentialsInvalid('Invalid Credentials')

    async def handle(self, request):
        async with self.auth_lock:
            try:
                await self.authenticate(request.headers)
            except AuthenticationInvalidOrMissing:
                return web.Response(headers={"WWW-Authenticate": "Basic realm=Electrum"},
                                    text='Unauthorized', status=401)
            except AuthenticationCredentialsInvalid:
                return web.Response(text='Forbidden', status=403)
        try:
            request = await request.text()
            request = json.loads(request)
            method = request['method']
            _id = request['id']
            params = request.get('params', [])  # type: Union[Sequence, Mapping]
            if method not in self._methods:
                raise Exception(f"attempting to use unregistered method: {method}")
            f = self._methods[method]
        except Exception as e:
            self.logger.exception("invalid request")
            return web.Response(text='Invalid Request', status=500)
        response = {
            'id': _id,
            'jsonrpc': '2.0',
        }
        try:
            if isinstance(params, dict):
                response['result'] = await f(**params)
            else:
                response['result'] = await f(*params)
        except UserFacingException as e:
            response['error'] = {
                'code': JsonRPCError.Codes.USERFACING,
                'message': str(e),
            }
        except BaseException as e:
            self.logger.exception("internal error while executing RPC")
            response['error'] = {
                'code': JsonRPCError.Codes.INTERNAL,
                'message': "internal error while executing RPC",
                'data': {
                    "exception": repr(e),
                    "traceback": "".join(traceback.format_exception(e)),
                },
            }
        return web.json_response(response)


class CommandsServer(AuthenticatedServer):

    def __init__(self, daemon: 'Daemon', fd, *, only_minimal_jsonrpc: bool):
        rpc_user, rpc_password = get_rpc_credentials(daemon.config)
        AuthenticatedServer.__init__(self, rpc_user, rpc_password)
        self.daemon = daemon
        self.fd = fd
        self._only_minimal_jsonrpc = only_minimal_jsonrpc
        self.config = daemon.config
        sockettype = self.config.RPC_SOCKET_TYPE
        self.socktype = sockettype if sockettype != 'auto' else get_rpcsock_default_type(self.config)
        self.sockpath = self.config.RPC_SOCKET_FILEPATH or get_rpcsock_defaultpath(self.config)
        self.host = self.config.RPC_HOST
        self.port = self.config.RPC_PORT
        self.cmd_runner = Commands(config=self.config, network=self.daemon.network, daemon=self.daemon)
        self.app = web.Application()
        self.app.router.add_post("/", self.handle)
        # First add always-enabled commands that are also available for "minimal" rpc server.
        # - "ping" RPC is needed for the lockfile fd to work.
        self.register_method('ping', self.ping)
        # - "gui" RPC is needed for URI handling. (TODO restrict further: disallow opening arbitrary file paths)
        self.register_method('gui', self.gui)
        # Add other commands:
        if not only_minimal_jsonrpc:
            for cmdname in known_commands:
                self.register_method(cmdname, getattr(self.cmd_runner, cmdname))
            self.register_method('run_cmdline', self.run_cmdline)

    def _socket_config_str(self) -> str:
        if self.socktype == 'unix':
            return f"<socket type={self.socktype}, path={self.sockpath}>"
        elif self.socktype == 'tcp':
            return f"<socket type={self.socktype}, host={self.host}, port={self.port}>"
        else:
            raise Exception(f"unknown socktype '{self.socktype!r}'")

    async def run(self):
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        if self.socktype == 'unix':
            site = web.UnixSite(self.runner, self.sockpath)
        elif self.socktype == 'tcp':
            site = web.TCPSite(self.runner, self.host, self.port)
        else:
            raise Exception(f"unknown socktype '{self.socktype!r}'")
        try:
            await site.start()
        except Exception as e:
            raise Exception(f"failed to start CommandsServer at {self._socket_config_str()}. got exc: {e!r}") from None
        # now server has started.
        if self.socktype == 'unix':
            # set restrictive permissions on unix domain socket.
            # FIXME race? we are late. should set this during socket-file creation but aiohttp API does not let us.
            os_chmod(self.sockpath, stat.S_IREAD | stat.S_IWRITE)
        # write server conn details into lockfile fd
        if self.socktype == 'unix':
            addr = self.sockpath
        elif self.socktype == 'tcp':
            socket = site._server.sockets[0]
            addr = socket.getsockname()
        else:
            raise Exception(f"impossible socktype ({self.socktype!r})")
        os.write(self.fd, bytes(repr((self.socktype, addr, time.time())), 'utf8'))
        os.close(self.fd)
        self.logger.info(
            f"now running and listening. socktype={self.socktype}, addr={addr}. "
            f"only_minimal_jsonrpc={self._only_minimal_jsonrpc}")

    async def ping(self):
        return True

    async def gui(self, config_options):
        # note: "config_options" is coming from the short-lived CLI-invocation,
        #        while self.config is the config of the long-lived daemon process.
        #       "config_options" should have priority.
        if self.daemon.gui_object:
            if hasattr(self.daemon.gui_object, 'new_window'):
                if config_options.get(SimpleConfig.NETWORK_OFFLINE.key()) and not self.config.NETWORK_OFFLINE:
                    raise UserFacingException(
                        "error: current GUI is running online, so it cannot open a new wallet offline.")
                path = config_options.get('wallet_path') or self.config.get_wallet_path()
                self.daemon.gui_object.new_window(path, config_options.get('url'))
                return True
            else:
                raise UserFacingException("error: current GUI does not support multiple windows")
        else:
            raise UserFacingException("error: Electrum is running in daemon mode. Please stop the daemon first.")

    async def run_cmdline(self, config_options):
        cmdname = config_options['cmd']
        cmd = known_commands.get(cmdname)
        if not cmd:
            return f"unknown command: {cmdname}"
        # arguments passed to function
        args = [config_options.get(x) for x in cmd.params]
        # decode json arguments
        args = [json_decode(i) for i in args]
        # options
        kwargs = {}
        for x in cmd.options:
            kwargs[x] = config_options.get(x)
        if 'wallet_path' in cmd.options or 'wallet' in cmd.options:
            wallet_path = config_options.get('wallet_path')
            if len(self.daemon._wallets) > 1 and wallet_path is None:
                raise UserFacingException("error: wallet not specified")
            kwargs['wallet_path'] = wallet_path
        func = getattr(self.cmd_runner, cmd.name)
        # execute requested command now.  note: cmd can raise, the caller (self.handle) will wrap it.
        result = await func(*args, **kwargs)
        return result


class Daemon(Logger):

    network: Optional[Network] = None
    gui_object: Optional['gui.BaseElectrumGui'] = None

    @profiler
    def __init__(
        self,
        config: SimpleConfig,
        fd=None,
        *,
        listen_jsonrpc: bool = True,
        only_minimal_jsonrpc: bool = True,
        start_network: bool = True,  # setting to False allows customising network settings before starting it
    ):
        Logger.__init__(self)
        self.config = config
        self.listen_jsonrpc = listen_jsonrpc
        if fd is None and listen_jsonrpc:
            fd = get_file_descriptor(config)
            if fd is None:
                raise Exception('failed to lock daemon; already running?')
        self._plugins = None  # type: Optional[Plugins]
        self.asyncio_loop = util.get_asyncio_loop()
        if not self.config.NETWORK_OFFLINE:
            self.network = Network(config, daemon=self)
        self.fx = FxThread(config=config)
        # wallet_key -> wallet
        self._wallets = {}  # type: Dict[str, Abstract_Wallet]
        self._wallet_lock = threading.RLock()

        self._stop_entered = False
        self._stopping_soon_or_errored = threading.Event()
        self._stopped_event = threading.Event()

        self.taskgroup = OldTaskGroup()
        asyncio.run_coroutine_threadsafe(self._run(), self.asyncio_loop)
        if start_network and self.network:
            self.start_network()
        # Setup commands server
        self.commands_server = None
        if listen_jsonrpc:
            self.commands_server = CommandsServer(self, fd, only_minimal_jsonrpc=only_minimal_jsonrpc)
            asyncio.run_coroutine_threadsafe(self.taskgroup.spawn(self.commands_server.run()), self.asyncio_loop)

    @log_exceptions
    async def _run(self):
        self.logger.info("starting taskgroup.")
        try:
            async with self.taskgroup as group:
                await group.spawn(asyncio.Event().wait)  # run forever (until cancel)
        except Exception as e:
            self.logger.exception("taskgroup died.")
            util.send_exception_to_crash_reporter(e)
        finally:
            self.logger.info("taskgroup stopped.")
            # note: we could just "await self.stop()", but in that case GUI users would
            #       not see the exception (especially if the GUI did not start yet).
            self._stopping_soon_or_errored.set()

    def start_network(self):
        self.logger.info(f"starting network.")
        assert not self.config.NETWORK_OFFLINE
        assert self.network
        self.network.start(jobs=[self.fx.run])
        # prepare lightning functionality, also load channel db early
        if self.config.LIGHTNING_USE_GOSSIP:
            self.network.start_gossip()

    @staticmethod
    def _wallet_key_from_path(path) -> str:
        """This does stricter path standardization than 'standardize_path'.
        It is used for keying the _wallets dict,
        but MUST NOT be used as a *path* for the actual filesystem operations. (see #8495)
        """
        path = standardize_path(path)
        # The extra normalisation makes it even harder to open the same wallet file multiple times simultaneously.
        # - "realpath" resolves symlinks:
        #   note: the path returned by realpath has been observed NOT to work for FS operations!
        #         (e.g. for Cryptomator WinFSP/FUSE mounts, see #8495).
        #         It is okay for us to use it for computing a canonical wallet *key*, but cannot be used as a path!
        try:
            path = os.path.realpath(path, strict=False)
        except OSError as e:  # see #10182
            _logger.warning(f"could not parse {path!r}: {e!r}")
            path = path
        # - "normcase" does Windows-specific case and slash normalisation:
        path = os.path.normcase(path)
        # - prepend header to break usage of wallet keys as fs paths
        header = "WALLETKEY-"
        return header + str(path)

    def with_wallet_lock(func):
        def func_wrapper(self: 'Daemon', *args, **kwargs):
            with self._wallet_lock:
                return func(self, *args, **kwargs)
        return func_wrapper

    @with_wallet_lock
    def load_wallet(
        self,
        path,
        password: Optional[str],
        *,
        upgrade: bool = False,
        force_check_password: bool = False,
    ) -> Optional[Abstract_Wallet]:
        """
        force_check_password: if False, the password arg is only used if it needed to decrypt the storage.
                              if True, the password arg is always validated.
        """
        assert password != ''
        path = standardize_path(path)
        wallet_key = self._wallet_key_from_path(path)
        # wizard will be launched if we return
        if wallet := self._wallets.get(wallet_key):
            if force_check_password:
                wallet.check_password(password)
            if self.config.get('wallet_path') is None:
                self.config.CURRENT_WALLET = path
            return wallet
        wallet = self._load_wallet(
            path, password, upgrade=upgrade, config=self.config, force_check_password=force_check_password)
        if self.network:
            wallet.start_network(self.network)
        elif wallet.lnworker:
            # in offline mode, we need to trigger callbacks
            coro = wallet.lnworker.lnwatcher.trigger_callbacks(requires_synchronizer=False)
            asyncio.run_coroutine_threadsafe(coro, self.asyncio_loop)
        self.add_wallet(wallet)
        if self.config.get('wallet_path') is None:
            self.config.CURRENT_WALLET = path
        self.update_recently_opened_wallets(path)
        return wallet


    @staticmethod
    @profiler
    def _load_wallet(
            path,
            password: Optional[str],
            *,
            upgrade: bool = False,
            config: SimpleConfig,
            force_check_password: bool = False,  # if set, always validate password
    ) -> Optional[Abstract_Wallet]:
        path = standardize_path(path)
        storage = WalletStorage(path, allow_partial_writes=config.WALLET_PARTIAL_WRITES)
        if not storage.file_exists():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)
        if storage.is_encrypted():
            if not password:
                raise InvalidPassword('No password given')
            storage.decrypt(password)
        # read data, pass it to db
        db = WalletDB(storage.read(), storage=storage, upgrade=upgrade)
        if db.get_action():
            raise WalletUnfinished(db)
        wallet = Wallet(db, config=config)
        if force_check_password:
            wallet.check_password(password)
        return wallet

    @with_wallet_lock
    def add_wallet(self, wallet: Abstract_Wallet) -> None:
        path = wallet.storage.get_path()
        wallet_key = self._wallet_key_from_path(path)
        self._wallets[wallet_key] = wallet
        run_hook('daemon_wallet_loaded', self, wallet)

    def get_wallet(self, path: str) -> Optional[Abstract_Wallet]:
        wallet_key = self._wallet_key_from_path(path)
        return self._wallets.get(wallet_key)

    @with_wallet_lock
    def get_wallets(self) -> Dict[str, Abstract_Wallet]:
        return dict(self._wallets)  # copy

    def delete_wallet(self, path: str) -> bool:
        self.stop_wallet(path)
        if os.path.exists(path):
            os.unlink(path)
            self.update_recently_opened_wallets(path, remove=True)
            if self.config.CURRENT_WALLET == path:
                self.config.CURRENT_WALLET = None
            return True
        return False

    def rename_wallet_file(self, old_path: str, new_path: str):
        old_path = standardize_path(old_path)
        new_path = standardize_path(new_path)
        if os.path.exists(new_path):
            raise ValueError("Wallet file already exists")
        os.rename(old_path, new_path)
        self.logger.debug(f'renamed wallet: {old_path} -> {new_path}')
        self.update_recently_opened_wallets(old_path, remove=True)
        if self.config.CURRENT_WALLET == old_path:
            self.config.CURRENT_WALLET = new_path

    def stop_wallet(self, path: str) -> bool:
        """Returns True iff a wallet was found."""
        assert util.get_running_loop() != util.get_asyncio_loop(), 'must not be called from asyncio thread'
        fut = asyncio.run_coroutine_threadsafe(self._stop_wallet(path), self.asyncio_loop)
        return fut.result()

    @with_wallet_lock
    async def _stop_wallet(self, path: str) -> bool:
        """Returns True iff a wallet was found."""
        path = standardize_path(path)
        wallet_key = self._wallet_key_from_path(path)
        wallet = self._wallets.pop(wallet_key, None)
        if not wallet:
            return False
        await wallet.stop()
        if self.config.get('wallet_path') is None:
            wallet_paths = [w.storage.get_path() for w in self._wallets.values()
                            if w.storage and w.storage.get_path()]
            if self.config.CURRENT_WALLET == path and wallet_paths:
                self.config.CURRENT_WALLET = wallet_paths[0]
        return True

    def run_daemon(self):
        if 'wallet_path' in self.config.cmdline_options:
            self.logger.warning("Ignoring parameter 'wallet_path' for daemon. "
                                "Use the load_wallet command instead.")
        # init plugins
        self._plugins = Plugins(self.config, 'cmdline')
        # block until we are stopping
        try:
            self._stopping_soon_or_errored.wait()
        except KeyboardInterrupt:
            self.logger.info("got KeyboardInterrupt")
        # we either initiate shutdown now,
        # or it has already been initiated (in which case this is a no-op):
        self.logger.info("run_daemon is calling stop()")
        asyncio.run_coroutine_threadsafe(self.stop(), self.asyncio_loop).result()
        # wait until "stop" finishes:
        self._stopped_event.wait()

    async def stop(self):
        if self._stop_entered:
            return
        self._stop_entered = True
        self._stopping_soon_or_errored.set()
        self.logger.info("stop() entered. initiating shutdown")
        try:
            if self.gui_object:
                self.gui_object.stop()
            self.logger.info("stopping all wallets")
            async with OldTaskGroup() as group:
                for k, wallet in self._wallets.items():
                    await group.spawn(wallet.stop())
            self.logger.info("stopping network and taskgroup")
            async with ignore_after(2):
                async with OldTaskGroup() as group:
                    if self.network:
                        await group.spawn(self.network.stop(full_shutdown=True))
                    await group.spawn(self.taskgroup.cancel_remaining())
            if self._plugins:
                self.logger.info("stopping plugins")
                self._plugins.stop()
                async with ignore_after(1):
                    await self._plugins.stopped_event_async.wait()
        finally:
            if self.listen_jsonrpc:
                self.logger.info("removing lockfile")
                remove_lockfile(get_lockfile(self.config))
            self.logger.info("stopped")
            self._stopped_event.set()

    def run_gui(self) -> None:
        assert self.config
        threading.current_thread().name = 'GUI'
        gui_name = self.config.GUI_NAME
        if gui_name in ['lite', 'classic']:
            gui_name = 'qt'
        self._plugins = Plugins(self.config, gui_name)  # init plugins
        self.logger.info(f'launching GUI: {gui_name}')
        try:
            try:
                gui = __import__('electrum.gui.' + gui_name, fromlist=['electrum'])
            except GuiImportError as e:
                sys.exit(str(e))
            self.gui_object = gui.ElectrumGui(config=self.config, daemon=self, plugins=self._plugins)
            if not self._stop_entered:
                self.gui_object.main()
            else:
                # If daemon.stop() was called before gui_object got created, stop gui now.
                self.gui_object.stop()
        except BaseException as e:
            self.logger.error(f'GUI raised exception: {repr(e)}. shutting down.')
            raise
        finally:
            # app will exit now
            asyncio.run_coroutine_threadsafe(self.stop(), self.asyncio_loop).result()

    @with_wallet_lock
    def check_password_for_directory(self, *, old_password, new_password=None, wallet_dir: str) -> Tuple[bool, bool, list[str]]:
        """Checks password against all wallets (in dir), returns whether they can be unified and whether they are already.
        If new_password is not None, update all wallet passwords to new_password.
        """
        assert os.path.exists(wallet_dir), f"path {wallet_dir!r} does not exist"
        succeeded = []
        failed = []
        is_unified = True
        for filename in os.listdir(wallet_dir):
            path = os.path.join(wallet_dir, filename)
            path = standardize_path(path)
            if not os.path.isfile(path):
                continue
            wallet = self.get_wallet(path)
            # note: we only create a new wallet object if one was not loaded into the daemon already.
            #       This is to avoid having two wallet objects contending for the same file.
            #       Take care: this only works if the daemon knows about all wallet objects.
            #                  if other code already has created a Wallet() for a file but did not tell the daemon,
            #                  hard-to-understand bugs will follow...
            if wallet is None:
                try:
                    wallet = self._load_wallet(path, old_password, upgrade=True, config=self.config)
                except util.InvalidPassword:
                    pass
                except Exception:
                    self.logger.exception(f'failed to load wallet at {path!r}:')
            if wallet is None:
                failed.append(path)
                continue
            if not wallet.storage.is_encrypted():
                is_unified = False
            try:
                try:
                    wallet.check_password(old_password)
                    old_password_real = old_password
                except util.InvalidPassword:
                    wallet.check_password(None)
                    old_password_real = None
            except Exception:
                failed.append(path)
                continue
            if new_password:
                self.logger.info(f'updating password for wallet: {path!r}')
                wallet.update_password(old_password_real, new_password, encrypt_storage=True)
            succeeded.append(path)

        can_be_unified = failed == []
        is_unified = can_be_unified and is_unified
        return can_be_unified, is_unified, succeeded

    @with_wallet_lock
    def update_password_for_directory(
            self,
            *,
            old_password,
            new_password,
            wallet_dir: Optional[str] = None,
    ) -> bool:
        """returns whether password is unified"""
        if new_password is None:
            # we opened a non-encrypted wallet
            return False
        if wallet_dir is None:
            wallet_dir = os.path.dirname(self.config.get_wallet_path())
        can_be_unified, is_unified, _ = self.check_password_for_directory(
            old_password=old_password, new_password=None, wallet_dir=wallet_dir)
        if not can_be_unified:
            return False
        if is_unified and old_password == new_password:
            return True
        self.check_password_for_directory(
            old_password=old_password, new_password=new_password, wallet_dir=wallet_dir)
        return True

    def update_recently_opened_wallets(self, wallet_path, *, remove: bool = False) -> None:
        if util.is_hidden_wallet_path(wallet_path):
            return None  # don't save "hidden wallet" paths
        recent = self.config.RECENTLY_OPEN_WALLET_FILES or []
        if wallet_path in recent:
            recent.remove(wallet_path)
        if not remove:
            recent.insert(0, wallet_path)
            recent = [path for path in recent if os.path.exists(path)]
            recent = recent[:5]
        self.config.RECENTLY_OPEN_WALLET_FILES = recent
        util.trigger_callback('recently_opened_wallets_update')



from electrum.plugin import hook
from electrum.util import print_msg, raw_input, print_stderr
from electrum.logging import get_logger

from electrum.hw_wallet.cmdline import CmdLineHandler

from .coldcard import ColdcardPlugin


_logger = get_logger(__name__)


class ColdcardCmdLineHandler(CmdLineHandler):

    def get_passphrase(self, msg, confirm):
        raise NotImplementedError

    def get_pin(self, msg, *, show_strength=True):
        raise NotImplementedError

    def prompt_auth(self, msg):
        raise NotImplementedError

    def yes_no_question(self, msg):
        print_msg(msg)
        return raw_input() in 'yY'

    def stop(self):
        pass

    def update_status(self, b):
        _logger.info(f'hw device status {b}')

    def finished(self):
        pass

class Plugin(ColdcardPlugin):
    handler = ColdcardCmdLineHandler()

    @hook
    def init_keystore(self, keystore):
        if not isinstance(keystore, self.keystore_class):
            return
        keystore.handler = self.handler

    def create_handler(self, window):
        return self.handler

# EOF



from abc import ABC, ABCMeta
import os.path
import time
import sys
import platform
import queue
import os
import webbrowser
import ctypes
from functools import partial, lru_cache
from typing import (NamedTuple, Callable, Optional, TYPE_CHECKING, List, Any, Sequence, Tuple, Union)

from PyQt6 import QtCore
from PyQt6.QtGui import (QFont, QColor, QCursor, QPixmap, QImage,
                         QPalette, QIcon, QFontMetrics, QPainter, QContextMenuEvent, QMovie)
from PyQt6.QtCore import (Qt, pyqtSignal, QCoreApplication, QSize, QRect, QPoint, QObject)
from PyQt6.QtWidgets import (QPushButton, QLabel, QMessageBox, QHBoxLayout, QVBoxLayout, QLineEdit,
                             QStyle, QDialog, QGroupBox, QButtonGroup, QRadioButton,
                             QFileDialog, QWidget, QToolButton, QPlainTextEdit, QApplication, QToolTip,
                             QGraphicsEffect, QGraphicsScene, QGraphicsPixmapItem, QLayoutItem, QLayout, QMenu,
                             QFrame, QAbstractButton)

from electrum.i18n import _
from electrum.util import (FileImportFailed, FileExportFailed, resource_path, EventListener,
                           get_logger, UserCancelled, UserFacingException, ChoiceItem)
from electrum.invoices import (PR_UNPAID, PR_PAID, PR_EXPIRED, PR_INFLIGHT, PR_UNKNOWN, PR_FAILED, PR_ROUTING,
                               PR_UNCONFIRMED, PR_BROADCASTING, PR_BROADCAST)
from electrum.qrreader import MissingQrDetectionLib, QrCodeResult
from electrum.submarine_swaps import pubkey_to_rgb_color

from electrum.gui.common_qt.util import TaskThread

if TYPE_CHECKING:
    from .main_window import ElectrumWindow
    from .paytoedit import PayToEdit

    from electrum.simple_config import SimpleConfig
    from electrum.simple_config import ConfigVarWithConfig


if platform.system() == 'Windows':
    MONOSPACE_FONT = 'Lucida Console'
elif platform.system() == 'Darwin':
    MONOSPACE_FONT = 'Monaco'
else:
    MONOSPACE_FONT = 'monospace'


_logger = get_logger(__name__)

dialogs = []

pr_icons = {
    PR_UNKNOWN: "warning.png",
    PR_UNPAID: "unpaid.png",
    PR_PAID: "confirmed.png",
    PR_EXPIRED: "expired.png",
    PR_INFLIGHT: "unconfirmed.png",
    PR_FAILED: "warning.png",
    PR_ROUTING: "unconfirmed.png",
    PR_UNCONFIRMED: "unconfirmed.png",
    PR_BROADCASTING: "unconfirmed.png",
    PR_BROADCAST: "unconfirmed.png",
}


# filter tx files in QFileDialog:
TRANSACTION_FILE_EXTENSION_FILTER_ANY = "Transaction (*.txn *.psbt);;All files (*)"
TRANSACTION_FILE_EXTENSION_FILTER_ONLY_PARTIAL_TX = "Partial Transaction (*.psbt)"
TRANSACTION_FILE_EXTENSION_FILTER_ONLY_COMPLETE_TX = "Complete Transaction (*.txn)"
TRANSACTION_FILE_EXTENSION_FILTER_SEPARATE = (f"{TRANSACTION_FILE_EXTENSION_FILTER_ONLY_PARTIAL_TX};;"
                                              f"{TRANSACTION_FILE_EXTENSION_FILTER_ONLY_COMPLETE_TX};;"
                                              f"All files (*)")


class EnterButton(QPushButton):
    def __init__(self, text, func):
        QPushButton.__init__(self, text)
        self.func = func
        self.clicked.connect(func)
        self._orig_text = text

    def keyPressEvent(self, e):
        if e.key() in [Qt.Key.Key_Return, Qt.Key.Key_Enter]:
            self.func()

    def restore_original_text(self):
        self.setText(self._orig_text)


class ThreadedButton(QPushButton):
    def __init__(self, text, task, on_success=None, on_error=None):
        QPushButton.__init__(self, text)
        self.task = task
        self.on_success = on_success
        self.on_error = on_error
        self.clicked.connect(self.run_task)

    def run_task(self):
        self.setEnabled(False)
        self.thread = TaskThread(self)
        self.thread.add(self.task, self.on_success, self.done, self.on_error)

    def done(self):
        self.setEnabled(True)
        self.thread.stop()


class WWLabel(QLabel):
    """Word-wrapping label"""
    def __init__(self, text="", parent=None):
        QLabel.__init__(self, text, parent)
        self.setWordWrap(True)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)


class RichLabel(WWLabel):
    """Word-wrapping label with link activation"""
    def __init__(self, text='', parent=None):
        WWLabel.__init__(self, text, parent)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        self.setOpenExternalLinks(True)


class AmountLabel(QLabel):
    def __init__(self, *args, **kwargs):
        QLabel.__init__(self, *args, **kwargs)
        self.setFont(QFont(MONOSPACE_FONT))
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)


class Spinner(QLabel):
    def __init__(self, *args, **kwargs):
        QLabel.__init__(self, *args, **kwargs)
        self.spinner = QMovie(icon_path('spinner.gif'))
        self.spinner.setScaledSize(QSize(20, 20))
        self.spinner.frameChanged.connect(lambda: self.setPixmap(self.spinner.currentPixmap()))
        self.setVisible(False)

    def setVisible(self, visible):
        if visible:
            self.spinner.start()
        else:
            self.spinner.stop()
        super().setVisible(visible)


class HelpMixin:
    def __init__(self, help_text: str, *, help_title: str = None):
        assert isinstance(self, QWidget), "HelpMixin must be a QWidget instance!"
        self.help_text = help_text
        self._help_title = help_title or _('Help')
        if isinstance(self, QLabel):
            self.setTextInteractionFlags(
                (self.textInteractionFlags() | Qt.TextInteractionFlag.TextSelectableByMouse)
                & ~Qt.TextInteractionFlag.TextSelectableByKeyboard)

    def show_help(self):
        custom_message_box(
            icon=QMessageBox.Icon.Information,
            parent=self,
            title=self._help_title,
            text=self.help_text,
            rich_text=True,
        )


class HelpLabel(HelpMixin, QLabel):

    def __init__(self, text: str, help_text: str):
        QLabel.__init__(self, text)
        HelpMixin.__init__(self, help_text)
        self.app = QCoreApplication.instance()
        self.font = self.font()

    @classmethod
    def from_configvar(cls, cv: 'ConfigVarWithConfig') -> 'HelpLabel':
        return HelpLabel(cv.get_short_desc() + ':', cv.get_long_desc())

    def mouseReleaseEvent(self, x):
        self.show_help()

    def enterEvent(self, event):
        self.font.setUnderline(True)
        self.setFont(self.font)
        self.app.setOverrideCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        return QLabel.enterEvent(self, event)

    def leaveEvent(self, event):
        self.font.setUnderline(False)
        self.setFont(self.font)
        self.app.setOverrideCursor(QCursor(Qt.CursorShape.ArrowCursor))
        return QLabel.leaveEvent(self, event)


class HelpButton(HelpMixin, QToolButton):
    def __init__(self, text: str):
        QToolButton.__init__(self)
        HelpMixin.__init__(self, text)
        self.setText('?')
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setFixedWidth(round(2.2 * char_width_in_lineedit()))
        self.clicked.connect(self.show_help)


class InfoButton(HelpMixin, QPushButton):
    def __init__(self, text: str):
        QPushButton.__init__(self, _('Info'))
        HelpMixin.__init__(self, text, help_title=_('Info'))
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setFixedWidth(6 * char_width_in_lineedit())
        self.clicked.connect(self.show_help)


class Buttons(QHBoxLayout):
    def __init__(self, *buttons):
        QHBoxLayout.__init__(self)
        self.addStretch(1)
        for b in buttons:
            if b is None:
                continue
            self.addWidget(b)


class CloseButton(QPushButton):
    def __init__(self, dialog):
        QPushButton.__init__(self, _("Close"))
        self.clicked.connect(dialog.close)
        self.setDefault(True)


class CopyButton(QPushButton):
    def __init__(self, text_getter, app):
        QPushButton.__init__(self, _("Copy"))
        self.clicked.connect(lambda: app.clipboard().setText(text_getter()))


class CopyCloseButton(QPushButton):
    def __init__(self, text_getter, app, dialog):
        QPushButton.__init__(self, _("Copy and Close"))
        self.clicked.connect(lambda: app.clipboard().setText(text_getter()))
        self.clicked.connect(dialog.close)
        self.setDefault(True)


class OkButton(QPushButton):
    def __init__(self, dialog, label=None):
        QPushButton.__init__(self, label or _("OK"))
        self.clicked.connect(dialog.accept)
        self.setDefault(True)


class CancelButton(QPushButton):
    def __init__(self, dialog, label=None):
        QPushButton.__init__(self, label or _("Cancel"))
        self.clicked.connect(dialog.reject)


class MessageBoxMixin(object):
    def top_level_window_recurse(self, window=None, test_func=None):
        window = window or self
        classes = (WindowModalDialog, QMessageBox)
        if test_func is None:
            test_func = lambda x: True
        for n, child in enumerate(window.children()):
            # Test for visibility as old closed dialogs may not be GC-ed.
            # Only accept children that confirm to test_func.
            if isinstance(child, classes) and child.isVisible() \
                    and test_func(child):
                return self.top_level_window_recurse(child, test_func=test_func)
        return window

    def top_level_window(self, test_func=None):
        return self.top_level_window_recurse(test_func)

    def question(self, msg, parent=None, title=None, icon=None, **kwargs) -> bool:
        yes, no = QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No
        return yes == self.msg_box(icon=icon or QMessageBox.Icon.Question,
                                   parent=parent,
                                   title=title or '',
                                   text=msg,
                                   buttons=yes | no,
                                   defaultButton=no,
                                   **kwargs)

    def show_warning(self, msg, parent=None, title=None, **kwargs):
        return self.msg_box(QMessageBox.Icon.Warning, parent,
                            title or _('Warning'), msg, **kwargs)

    def show_error(self, msg, parent=None, **kwargs):
        return self.msg_box(QMessageBox.Icon.Warning, parent,
                            _('Error'), msg, **kwargs)

    def show_critical(self, msg, parent=None, title=None, **kwargs):
        return self.msg_box(QMessageBox.Icon.Critical, parent,
                            title or _('Critical Error'), msg, **kwargs)

    def show_message(self, msg, parent=None, title=None, icon=QMessageBox.Icon.Information, **kwargs):
        return self.msg_box(icon, parent, title or _('Information'), msg, **kwargs)

    def msg_box(
            self,
            icon: Union[QMessageBox.Icon, QPixmap],
            parent: QWidget,
            title: str,
            text: str,
            *,
            buttons: Union[QMessageBox.StandardButton,
                           List[Union[QMessageBox.StandardButton, Tuple[QAbstractButton, QMessageBox.ButtonRole, int]]]] = QMessageBox.StandardButton.Ok,
            defaultButton: QMessageBox.StandardButton = QMessageBox.StandardButton.NoButton,
            rich_text: bool = False,
            checkbox: Optional[bool] = None
    ):
        parent = parent or self.top_level_window()
        return custom_message_box(
            icon=icon, parent=parent, title=title, text=text, buttons=buttons, defaultButton=defaultButton,
            rich_text=rich_text, checkbox=checkbox
        )

    def query_choice(
        self,
        msg: Optional[str],
        choices: Sequence['ChoiceItem'],
        *,
        title: Optional[str] = None,
        default_key: Optional[Any] = None,
    ) -> Optional[Any]:
        """Returns ChoiceItem.key (for selected item), or None if the user cancels the dialog.

        Needed by QtHandler for hardware wallets.
        """
        if title is None:
            title = _('Question')
        dialog = WindowModalDialog(self.top_level_window(), title=title)
        dialog.setMinimumWidth(400)
        choice_widget = ChoiceWidget(message=msg, choices=choices, default_key=default_key)
        vbox = QVBoxLayout(dialog)
        vbox.addWidget(choice_widget)
        cancel_button = CancelButton(dialog)
        vbox.addLayout(Buttons(cancel_button, OkButton(dialog)))
        cancel_button.setFocus()
        if not dialog.exec():
            return None
        return choice_widget.selected_key

    def password_dialog(self, msg=None, parent=None):
        from .password_dialog import PasswordDialog
        parent = parent or self
        d = PasswordDialog(parent, msg)
        return d.run()


def custom_message_box(
        *,
        icon: Union[QMessageBox.Icon, QPixmap],
        parent: QWidget,
        title: str,
        text: str,
        buttons: Union[QMessageBox.StandardButton,
                       List[Union[QMessageBox.StandardButton, Tuple[QAbstractButton, QMessageBox.ButtonRole, int]]]] = QMessageBox.StandardButton.Ok,
        defaultButton: QMessageBox.StandardButton = QMessageBox.StandardButton.NoButton,
        rich_text: bool = False,
        checkbox: Optional[bool] = None
) -> int:
    custom_buttons = []
    standard_buttons = QMessageBox.StandardButton.NoButton
    if buttons:
        if not isinstance(buttons, list):
            buttons = [buttons]
        for button in buttons:
            if isinstance(button, QMessageBox.StandardButton):
                standard_buttons |= button
            else:
                custom_buttons.append(button)
    if type(icon) is QPixmap:
        d = QMessageBox(QMessageBox.Icon.Information, title, str(text), standard_buttons, parent)
        d.setIconPixmap(icon)
    else:
        d = QMessageBox(icon, title, str(text), standard_buttons, parent)
    for button, role, _ in custom_buttons:
        d.addButton(button, role)
    d.setWindowModality(Qt.WindowModality.WindowModal)
    d.setDefaultButton(defaultButton)
    if rich_text:
        d.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.LinksAccessibleByMouse)
        # set AutoText instead of RichText
        # AutoText lets Qt figure out whether to render as rich text.
        # e.g. if text is actually plain text and uses "\n" newlines;
        #      and we set RichText here, newlines would be swallowed
        d.setTextFormat(Qt.TextFormat.AutoText)
    else:
        d.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        d.setTextFormat(Qt.TextFormat.PlainText)
    if checkbox is not None:
        d.setCheckBox(checkbox)
    result = d.exec()
    for button, _, value in custom_buttons:
        if button == d.clickedButton():
            return value
    return result


class WindowModalDialog(QDialog, MessageBoxMixin):
    '''Handy wrapper; window modal dialogs are better for our multi-window
    daemon model as other wallet windows can still be accessed.'''
    def __init__(self, parent, title=None):
        QDialog.__init__(self, parent)
        self.setWindowModality(Qt.WindowModality.WindowModal)
        if title:
            self.setWindowTitle(title)


class WaitingDialog(WindowModalDialog):
    '''Shows a please wait dialog whilst running a task.  It is not
    necessary to maintain a reference to this dialog.'''
    def __init__(self, parent: QWidget, message: str, task, on_success=None, on_error=None, on_cancel=None):
        assert parent
        if isinstance(parent, MessageBoxMixin):
            parent = parent.top_level_window()
        WindowModalDialog.__init__(self, parent, _("Please wait"))
        self.message_label = QLabel(message)
        vbox = QVBoxLayout(self)
        vbox.addWidget(self.message_label)
        if on_cancel:
            self.cancel_button = CancelButton(self)
            self.cancel_button.clicked.connect(on_cancel)
            vbox.addLayout(Buttons(self.cancel_button))
        self.accepted.connect(self.on_accepted)
        self.show()
        self.thread = TaskThread(self)
        self.thread.finished.connect(self.deleteLater)  # see #3956
        self.thread.add(task, on_success, self.accept, on_error)

    def wait(self):
        self.thread.wait()

    def on_accepted(self):
        self.thread.stop()

    def update(self, msg):
        print(msg)
        self.message_label.setText(msg)


class RunCoroutineDialog(WaitingDialog):

    def __init__(self, parent: QWidget, message: str, coroutine):
        from electrum import util
        import asyncio
        import concurrent.futures
        loop = util.get_asyncio_loop()
        assert util.get_running_loop() != loop, 'must not be called from asyncio thread'
        self._exception = None
        self._result = None
        self._future = asyncio.run_coroutine_threadsafe(coroutine, loop)
        def task():
            try:
                self._result = self._future.result()
            except concurrent.futures.CancelledError:
                self._exception = UserCancelled
            except Exception as e:
                self._exception = e
        WaitingDialog.__init__(self, parent, message, task, on_cancel=self._future.cancel)

    def run(self):
        self.exec()
        if self._exception:
            raise self._exception
        else:
            return self._result


def line_dialog(parent, title, label, ok_label, default=None):
    dialog = WindowModalDialog(parent, title)
    dialog.setMinimumWidth(500)
    l = QVBoxLayout()
    dialog.setLayout(l)
    l.addWidget(QLabel(label))
    txt = QLineEdit()
    if default:
        txt.setText(default)
    l.addWidget(txt)
    l.addLayout(Buttons(CancelButton(dialog), OkButton(dialog, ok_label)))
    if dialog.exec():
        return txt.text()


def text_dialog(
        *,
        parent,
        title,
        header_layout,
        ok_label,
        default=None,
        allow_multi=False,
        config: 'SimpleConfig',
):
    from .qrtextedit import ScanQRTextEdit
    dialog = WindowModalDialog(parent, title)
    dialog.setMinimumWidth(600)
    l = QVBoxLayout()
    dialog.setLayout(l)
    if isinstance(header_layout, str):
        l.addWidget(QLabel(header_layout))
    else:
        l.addLayout(header_layout)
    txt = ScanQRTextEdit(allow_multi=allow_multi, config=config)
    if default:
        txt.setText(default)
    l.addWidget(txt)
    l.addLayout(Buttons(CancelButton(dialog), OkButton(dialog, ok_label)))
    if dialog.exec():
        return txt.toPlainText()


class ChoiceWidget(QWidget):
    """Renders a list of ChoiceItems as a radiobuttons group.
    Callers can pre-select an item by key, through the 'default_key' parameter.
    The selected item is made available by index (selected_index),
    by key (selected_key) and by Choice (selected_item).
    """

    itemSelected = pyqtSignal([int], arguments=['index'])

    def __init__(
        self,
        *,
        message: Optional[str] = None,
        choices: Sequence[ChoiceItem] = None,
        default_key: Optional[Any] = None,
    ):
        QWidget.__init__(self)
        vbox = QVBoxLayout()
        self.setLayout(vbox)

        if choices is None:
            choices = []

        self.selected_index = -1   # type: int
        self.selected_item = None  # type: Optional[ChoiceItem]
        self.selected_key = None   # type: Optional[Any]
        self.choices = choices     # type: Sequence[ChoiceItem]

        if message and len(message) > 50:
            vbox.addWidget(WWLabel(message))
            message = ""
        gb2 = QGroupBox(message)
        vbox.addWidget(gb2)
        vbox2 = QVBoxLayout()
        gb2.setLayout(vbox2)
        self.group = group = QButtonGroup()
        assert isinstance(choices, list)
        for i, c in enumerate(choices):
            assert isinstance(c, ChoiceItem), f"{c=!r}"
            button = QRadioButton(gb2)
            button.setText(c.label)
            vbox2.addWidget(button)
            group.addButton(button)
            group.setId(button, i)
            if (i == 0 and default_key is None) or c.key == default_key:
                self.selected_index = i
                self.selected_item = c
                self.selected_key = c.key
                button.setChecked(True)
        group.buttonClicked.connect(self.on_selected)

    def on_selected(self, button):
        self.selected_index = self.group.id(button)
        self.selected_item = self.choices[self.selected_index]
        self.selected_key = self.choices[self.selected_index].key
        self.itemSelected.emit(self.selected_index)

    def select(self, key):
        for i, c in enumerate(self.choices):
            if key == c.key:
                self.group.button(i).click()


class ResizableStackedWidget(QWidget):
    """Simple alternative to QStackedWidget, as QStackedWidget always resizes to the largest
       widget in the stack, leaving ugly scrollbars where they're not needed."""
    def __init__(self, parent):
        super().__init__(parent)
        self.setLayout(QVBoxLayout())
        self.widgets = []
        self.current_index = -1

    def sizeHint(self) -> QSize:
        if not self.count() or not self.currentWidget():
            return super().sizeHint()
        return self.currentWidget().sizeHint()

    def addWidget(self, widget: QWidget) -> int:
        self.widgets.append(widget)
        self.layout().addWidget(widget)
        if len(self.widgets) == 1:  # first widget?
            self.current_index = 0
        self.showCurrentWidget()
        return len(self.widgets) - 1

    def removeWidget(self, widget: QWidget):
        i = self.widgets.index(widget)
        self.widgets.remove(widget)
        self.layout().removeWidget(widget)
        if self.current_index >= i:
            self.current_index -= 1
            if self.current_index == self.count() - 1:
                self.showCurrentWidget()

    def setCurrentIndex(self, index: int):
        assert isinstance(index, int)
        assert 0 <= index < len(self.widgets), f'invalid widget index {index}'
        self.current_index = index
        self.showCurrentWidget()

    def currentWidget(self) -> Optional[QWidget]:
        if self.current_index < 0:
            return None
        return self.widgets[self.current_index]

    def showCurrentWidget(self):
        if not self.widgets:
            return

        for i, k in enumerate(self.widgets):
            if i == self.current_index:
                k.show()
            else:
                k.hide()

    def count(self) -> int:
        return len(self.widgets)


class VLine(QFrame):
    """Vertical line separator"""
    def __init__(self):
        super(VLine, self).__init__()
        self.setFrameShape(QFrame.Shape.VLine)
        self.setFrameShadow(QFrame.Shadow.Sunken)
        self.setLineWidth(1)


def address_field(addresses, *, btn_text: str = None):
    if btn_text is None:
        btn_text = _('Get wallet address')
    hbox = QHBoxLayout()
    address_e = QLineEdit()
    if addresses and len(addresses) > 0:
        address_e.setText(addresses[0])
    else:
        addresses = []

    def func():
        try:
            i = addresses.index(str(address_e.text())) + 1
            i = i % len(addresses)
            address_e.setText(addresses[i])
        except ValueError:
            # the user might have changed address_e to an
            # address not in the wallet (or to something that isn't an address)
            if addresses and len(addresses) > 0:
                address_e.setText(addresses[0])
    button = QPushButton(btn_text)
    button.clicked.connect(func)
    hbox.addWidget(button)
    hbox.addWidget(address_e)
    return hbox, address_e


def filename_field(parent, config, defaultname, select_msg):
    vbox = QVBoxLayout()
    vbox.addWidget(QLabel(_("Format")))
    gb = QGroupBox("format", parent)
    b1 = QRadioButton(gb)
    b1.setText(_("CSV"))
    b1.setChecked(True)
    b2 = QRadioButton(gb)
    b2.setText(_("json"))
    vbox.addWidget(b1)
    vbox.addWidget(b2)

    hbox = QHBoxLayout()

    directory = config.IO_DIRECTORY
    path = os.path.join(directory, defaultname)
    filename_e = QLineEdit()
    filename_e.setText(path)

    def func():
        text = filename_e.text()
        _filter = "*.csv" if defaultname.endswith(".csv") else "*.json" if defaultname.endswith(".json") else None
        p = getSaveFileName(
            parent=None,
            title=select_msg,
            filename=text,
            filter=_filter,
            config=config,
        )
        if p:
            filename_e.setText(p)

    button = QPushButton(_('File'))
    button.clicked.connect(func)
    hbox.addWidget(button)
    hbox.addWidget(filename_e)
    vbox.addLayout(hbox)

    def set_csv(v):
        text = filename_e.text()
        text = text.replace(".json",".csv") if v else text.replace(".csv",".json")
        filename_e.setText(text)

    b1.clicked.connect(lambda: set_csv(True))
    b2.clicked.connect(lambda: set_csv(False))

    return vbox, filename_e, b1


def get_icon_qrcode() -> QIcon:
    name = "qrcode_white.png" if ColorScheme.dark_scheme else "qrcode.png"
    return read_QIcon(name)


def get_icon_camera() -> QIcon:
    name = "camera_white.png" if ColorScheme.dark_scheme else "camera_dark.png"
    return read_QIcon(name)


def pubkey_to_q_icon(server_pubkey: str) -> QIcon:
    color = QColor(*pubkey_to_rgb_color(server_pubkey))
    color_pixmap = QPixmap(100, 100)
    color_pixmap.fill(color)
    return QIcon(color_pixmap)


def add_input_actions_to_context_menu(gih: 'GenericInputHandler', m: QMenu) -> None:
    if gih.on_qr_from_camera_input_btn:
        m.addAction(get_icon_camera(), _("Read QR code with camera"), gih.on_qr_from_camera_input_btn)
    if gih.on_qr_from_screenshot_input_btn:
        m.addAction(read_QIcon("picture_in_picture.png"), _("Read QR code from screen"), gih.on_qr_from_screenshot_input_btn)
    if gih.on_qr_from_file_input_btn:
        m.addAction(read_QIcon("qr_file.png"), _("Read QR code from file"), gih.on_qr_from_file_input_btn)
    if gih.on_input_file:
        m.addAction(read_QIcon("file.png"), _("Read text from file"), gih.on_input_file)


def scan_qr_from_screenshot() -> QrCodeResult:
    from .qrreader import scan_qr_from_image
    screenshots = [screen.grabWindow(0).toImage()
                   for screen in QApplication.instance().screens()]
    if all(screen.allGray() for screen in screenshots):
        raise UserFacingException(_("Failed to take screenshot."))
    scanned_qr = None
    for screenshot in screenshots:
        try:
            scan_result = scan_qr_from_image(screenshot)
        except MissingQrDetectionLib as e:
            raise UserFacingException(_("Unable to scan image.") + "\n" + repr(e))
        if len(scan_result) > 0:
            if (scanned_qr is not None) or len(scan_result) > 1:
                raise UserFacingException(_("More than one QR code was found on the screen."))
            scanned_qr = scan_result
    if scanned_qr is None:
        raise UserFacingException(_("No QR code was found on the screen."))
    assert len(scanned_qr) == 1, f"{len(scanned_qr)=}, expected 1"
    return scanned_qr[0]


class GenericInputHandler:
    on_qr_from_camera_input_btn: Callable[[], None] = None
    on_qr_from_screenshot_input_btn: Callable[[], None] = None
    on_qr_from_file_input_btn: Callable[[], None] = None
    on_input_file: Callable[[], None] = None

    def input_qr_from_camera(
            self,
            *,
            config: 'SimpleConfig',
            allow_multi: bool = False,
            show_error: Callable[[str], None],
            setText: Callable[[str], None] = None,
            parent: QWidget = None,
    ) -> None:
        if setText is None:
            setText = self.setText
        def cb(success: bool, error: str, data: Optional[str]):
            if not success:
                if error:
                    show_error(error)
                return
            if not data:
                data = ''
            try:
                if allow_multi:
                    text = self.text()
                    if data in text:
                        return
                    if text and not text.endswith('\n'):
                        text += '\n'
                    text += data
                    text += '\n'
                    setText(text)
                else:
                    new_text = data
                    setText(new_text)
            except Exception as e:
                show_error(_('Invalid payment identifier in QR') + ':\n' + repr(e))

        from .qrreader import scan_qrcode_from_camera
        if parent is None:
            parent = self if isinstance(self, QWidget) else None
        scan_qrcode_from_camera(parent=parent, config=config, callback=cb)

    def input_qr_from_screenshot(
            self,
            *,
            allow_multi: bool = False,
            show_error: Callable[[str], None],
            setText: Callable[[str], None] = None,
    ) -> None:
        if setText is None:
            setText = self.setText
        try:
            scanned_qr = scan_qr_from_screenshot()
        except UserFacingException as e:
            show_error(str(e))
            return
        data = scanned_qr.data
        try:
            if allow_multi:
                text = self.text()
                if data in text:
                    return
                if text and not text.endswith('\n'):
                    text += '\n'
                text += data
                text += '\n'
                setText(text)
            else:
                new_text = data
                setText(new_text)
        except Exception as e:
            show_error(_('Invalid payment identifier in QR') + ':\n' + repr(e))

    def input_file(
            self,
            *,
            config: 'SimpleConfig',
            show_error: Callable[[str], None],
            setText: Callable[[str], None] = None,
    ) -> None:
        if setText is None:
            setText = self.setText
        fileName = getOpenFileName(
            parent=None,
            title='select file',
            # trying to open non-text things like pdfs makes electrum freeze
            filter="Text files (*.txt *.csv);;All files (*)",
            config=config,
        )
        if not fileName:
            return
        try:
            try:
                with open(fileName, "r") as f:
                    data = f.read()
            except UnicodeError as e:
                with open(fileName, "rb") as f:
                    data = f.read()
                data = data.hex()
        except BaseException as e:
            show_error(_('Error opening file') + ':\n' + repr(e))
        else:
            try:
                setText(data)
            except Exception as e:
                show_error(_('Invalid payment identifier in file') + ':\n' + repr(e))

    def input_qr_from_file(
        self,
        *,
        allow_multi: bool = False,
        config: 'SimpleConfig',
        show_error: Callable[[str], None],
        setText: Callable[[str], None] = None,
    ):
        from .qrreader import scan_qr_from_image
        if setText is None:
            setText = self.setText

        file_name = getOpenFileName(
            parent=None,
            title=_("Select image file"),
            config=config,
            filter="Image files (*.png *.jpg *.jpeg *.bmp);;",
        )
        if not file_name:
            return
        image = QImage(file_name)
        if image.isNull():
            show_error(_("Failed to open image file."))
            return
        try:
            scan_result: Sequence[QrCodeResult] = scan_qr_from_image(image)
        except MissingQrDetectionLib as e:
            show_error(_("Unable to scan image.") + "\n" + repr(e))
            return
        if len(scan_result) < 1:
            show_error(_("No QR code was found in the image."))
            return
        if len(scan_result) > 1 and not allow_multi:
            show_error(_("More than one QR code was found in the image."))
            return

        if len(scan_result) > 1:
            result_text = "\n".join([r.data for r in scan_result])
        else:
            result_text = scan_result[0].data

        try:
            setText(result_text)
        except Exception as e:
            show_error(_("Couldn't set result") + ':\n' + repr(e))

    def input_paste_from_clipboard(
            self,
            *,
            setText: Callable[[str], None] = None,
    ) -> None:
        if setText is None:
            setText = self.setText
        app = QApplication.instance()
        setText(app.clipboard().text())


class OverlayControlMixin(GenericInputHandler):
    STYLE_SHEET_COMMON = '''
    QPushButton { border-width: 1px; padding: 0px; margin: 0px; }
    '''

    STYLE_SHEET_LIGHT = '''
    QPushButton { border: 1px solid transparent; }
    QPushButton:hover { border: 1px solid #3daee9; }
    '''

    def __init__(self, middle: bool = False):
        GenericInputHandler.__init__(self)
        assert isinstance(self, QWidget)
        assert isinstance(self, OverlayControlMixin)  # only here for type-hints in IDE
        self.middle = middle
        self.overlay_widget = QWidget(self)
        style_sheet = self.STYLE_SHEET_COMMON
        if not ColorScheme.dark_scheme:
            style_sheet = style_sheet + self.STYLE_SHEET_LIGHT
        self.overlay_widget.setStyleSheet(style_sheet)
        self.overlay_layout = QHBoxLayout(self.overlay_widget)
        self.overlay_layout.setContentsMargins(0, 0, 0, 0)
        self.overlay_layout.setSpacing(1)
        self._updateOverlayPos()

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self._updateOverlayPos()

    def _updateOverlayPos(self):
        frame_width = self.style().pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth)
        overlay_size = self.overlay_widget.sizeHint()
        x = self.rect().right() - frame_width - overlay_size.width()
        y = self.rect().bottom() - overlay_size.height()
        middle = self.middle
        if hasattr(self, 'document'):
            # Keep the buttons centered if we have less than 2 lines in the editor
            line_spacing = QFontMetrics(self.document().defaultFont()).lineSpacing()
            if self.rect().height() < (line_spacing * 2):
                middle = True
        y = (y / 2) + frame_width if middle else y - frame_width
        if hasattr(self, 'verticalScrollBar') and self.verticalScrollBar().isVisible():
            scrollbar_width = self.style().pixelMetric(QStyle.PixelMetric.PM_ScrollBarExtent)
            x -= scrollbar_width
        self.overlay_widget.move(int(x), int(y))

    def addWidget(self, widget: QWidget):
        # The old code positioned the items the other way around, so we just insert at position 0 instead
        self.overlay_layout.insertWidget(0, widget)

    def addButton(self, icon: QIcon, on_click, tooltip: str) -> QPushButton:
        button = QPushButton(self.overlay_widget)
        button.setToolTip(tooltip)
        button.setIcon(icon)
        button.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        button.clicked.connect(on_click)
        self.addWidget(button)
        return button

    def addCopyButton(self):
        def on_copy():
            app = QApplication.instance()
            app.clipboard().setText(self.text())
            QToolTip.showText(QCursor.pos(), _("Text copied to clipboard"), self)
        self.addButton(read_QIcon("copy.png"), on_copy, _("Copy to clipboard"))

    def addPasteButton(
            self,
            *,
            setText: Callable[[str], None] = None,
    ):
        input_paste_from_clipboard = partial(
            self.input_paste_from_clipboard,
            setText=setText,
        )
        self.addButton(read_QIcon("copy.png"), input_paste_from_clipboard, _("Paste from clipboard"))

    def add_qr_show_button(self, *, config: 'SimpleConfig', title: Optional[str] = None):
        if title is None:
            title = _("QR code")

        def qr_show():
            from .qrcodewidget import QRDialog
            try:
                s = str(self.text())
            except Exception:
                s = self.text()
            if not s:
                return
            QRDialog(
                data=s,
                parent=self,
                title=title,
                config=config,
            ).exec()

        self.addButton(get_icon_qrcode(), qr_show, _("Show as QR code"))
        # side-effect: we export this method:
        self.on_qr_show_btn = qr_show

    def add_qr_input_from_camera_button(
            self,
            *,
            config: 'SimpleConfig',
            allow_multi: bool = False,
            show_error: Callable[[str], None],
            setText: Callable[[str], None] = None,
    ):
        input_qr_from_camera = partial(
            self.input_qr_from_camera,
            config=config,
            allow_multi=allow_multi,
            show_error=show_error,
            setText=setText,
        )
        self.addButton(get_icon_camera(), input_qr_from_camera, _("Read QR code with camera"))
        # side-effect: we export these methods:
        self.on_qr_from_camera_input_btn = input_qr_from_camera

    def add_file_input_button(
            self,
            *,
            config: 'SimpleConfig',
            show_error: Callable[[str], None],
            setText: Callable[[str], None] = None,
    ) -> None:
        input_file = partial(
            self.input_file,
            config=config,
            show_error=show_error,
            setText=setText,
        )
        self.addButton(read_QIcon("file.png"), input_file, _("Read file"))

    def add_menu_button(
            self,
            *,
            options: Sequence[Tuple[Optional[Union[str, QIcon]], str, Callable[[], None]]],  # list of (icon, text, cb)
            icon: Optional[QIcon] = None,
            tooltip: Optional[str] = None,
    ):
        if icon is None:
            icon_name = "menu_vertical_white.png" if ColorScheme.dark_scheme else "menu_vertical.png"
            icon = read_QIcon(icon_name)
        if tooltip is None:
            tooltip = _("Other options")
        btn = self.addButton(icon, lambda: None, tooltip)
        menu = QMenu()
        for opt_icon, opt_text, opt_cb in options:
            if opt_icon is None:
                menu.addAction(opt_text, opt_cb)
            else:
                opt_icon = read_QIcon(opt_icon) if isinstance(opt_icon, str) else opt_icon
                menu.addAction(opt_icon, opt_text, opt_cb)
        btn.setMenu(menu)


class ButtonsLineEdit(OverlayControlMixin, QLineEdit):
    def __init__(self, text=None):
        QLineEdit.__init__(self, text)
        OverlayControlMixin.__init__(self, middle=True)


class ShowQRLineEdit(ButtonsLineEdit):
    """ read-only line with qr and copy buttons """
    def __init__(self, text: str, config, title=None):
        ButtonsLineEdit.__init__(self, text)
        self.setReadOnly(True)
        self.setFont(QFont(MONOSPACE_FONT))
        self.add_qr_show_button(config=config, title=title)
        self.addCopyButton()


class ButtonsTextEdit(OverlayControlMixin, QPlainTextEdit):
    def __init__(self, text=None):
        QPlainTextEdit.__init__(self, text)
        OverlayControlMixin.__init__(self)
        self.setText = self.setPlainText
        self.text = self.toPlainText


class PasswordLineEdit(QLineEdit):
    def __init__(self, *args, **kwargs):
        QLineEdit.__init__(self, *args, **kwargs)
        self.setEchoMode(QLineEdit.EchoMode.Password)

    def clear(self):
        # Try to actually overwrite the memory.
        # This is really just a best-effort thing...
        self.setText(len(self.text()) * " ")
        super().clear()


class ColorSchemeItem:
    def __init__(self, fg_color, bg_color):
        self.colors = (fg_color, bg_color)

    def _get_color(self, background):
        return self.colors[(int(background) + int(ColorScheme.dark_scheme)) % 2]

    def as_stylesheet(self, background=False):
        css_prefix = "background-" if background else ""
        color = self._get_color(background)
        return "QWidget {{ {}color:{}; }}".format(css_prefix, color)

    def as_color(self, background=False):
        color = self._get_color(background)
        return QColor(color)


class ColorScheme:
    dark_scheme = False

    GREEN = ColorSchemeItem("#117c11", "#8af296")
    YELLOW = ColorSchemeItem("#897b2a", "#ffff00")
    RED = ColorSchemeItem("#7c1111", "#f18c8c")
    BLUE = ColorSchemeItem("#123b7c", "#8cb3f2")
    LIGHTBLUE = ColorSchemeItem("black", "#d0f0ff")
    DEFAULT = ColorSchemeItem("black", "white")
    GRAY = ColorSchemeItem("gray", "gray")
    ORANGE = ColorSchemeItem("#ff9b45", "#ff9b45")

    @staticmethod
    def has_dark_background(widget):
        brightness = sum(widget.palette().color(QPalette.ColorRole.Window).getRgb()[0:3])
        return brightness < (255*3/2)

    @staticmethod
    def update_from_widget(widget, force_dark=False):
        ColorScheme.dark_scheme = bool(force_dark or ColorScheme.has_dark_background(widget))


class AcceptFileDragDrop:
    def __init__(self, file_type=""):
        assert isinstance(self, QWidget)
        self.setAcceptDrops(True)
        self.file_type = file_type

    def validateEvent(self, event):
        if not event.mimeData().hasUrls():
            event.ignore()
            return False
        for url in event.mimeData().urls():
            if not url.toLocalFile().endswith(self.file_type):
                event.ignore()
                return False
        event.accept()
        return True

    def dragEnterEvent(self, event):
        self.validateEvent(event)

    def dragMoveEvent(self, event):
        if self.validateEvent(event):
            event.setDropAction(Qt.DropAction.CopyAction)

    def dropEvent(self, event):
        if self.validateEvent(event):
            for url in event.mimeData().urls():
                self.onFileAdded(url.toLocalFile())

    def onFileAdded(self, fn):
        raise NotImplementedError()


def import_meta_gui(electrum_window: 'ElectrumWindow', title, importer, on_success):
    filter_ = "JSON (*.json);;All files (*)"
    filename = getOpenFileName(
        parent=electrum_window,
        title=_("Open {} file").format(title),
        filter=filter_,
        config=electrum_window.config,
    )
    if not filename:
        return
    try:
        importer(filename)
    except FileImportFailed as e:
        electrum_window.show_critical(str(e))
    else:
        electrum_window.show_message(_("Your {} were successfully imported").format(title))
        on_success()


def export_meta_gui(electrum_window: 'ElectrumWindow', title, exporter):
    filter_ = "JSON (*.json);;All files (*)"
    filename = getSaveFileName(
        parent=electrum_window,
        title=_("Select file to save your {}").format(title),
        filename='electrum_{}.json'.format(title),
        filter=filter_,
        config=electrum_window.config,
    )
    if not filename:
        return
    try:
        exporter(filename)
    except FileExportFailed as e:
        electrum_window.show_critical(str(e))
    else:
        electrum_window.show_message(_("Your {0} were exported to '{1}'")
                                     .format(title, str(filename)))


def getOpenFileName(*, parent, title, filter="", config: 'SimpleConfig') -> Optional[str]:
    """Custom wrapper for getOpenFileName that remembers the path selected by the user."""
    directory = config.IO_DIRECTORY
    fileName, __ = QFileDialog.getOpenFileName(parent, title, directory, filter)
    if fileName and directory != os.path.dirname(fileName):
        config.IO_DIRECTORY = os.path.dirname(fileName)
    return fileName


def getSaveFileName(
        *,
        parent,
        title,
        filename,
        filter="",
        default_extension: str = None,
        default_filter: str = None,
        config: 'SimpleConfig',
) -> Optional[str]:
    """Custom wrapper for getSaveFileName that remembers the path selected by the user."""
    directory = config.IO_DIRECTORY
    path = os.path.join(directory, filename)

    file_dialog = QFileDialog(parent, title, path, filter)
    file_dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
    if default_extension:
        # note: on MacOS, the selected filter's first extension seems to have priority over this...
        file_dialog.setDefaultSuffix(default_extension)
    if default_filter:
        assert default_filter in filter, f"default_filter={default_filter!r} does not appear in filter={filter!r}"
        file_dialog.selectNameFilter(default_filter)
    if file_dialog.exec() != QDialog.DialogCode.Accepted:
        return None

    selected_path = file_dialog.selectedFiles()[0]
    if selected_path and directory != os.path.dirname(selected_path):
        config.IO_DIRECTORY = os.path.dirname(selected_path)
    return selected_path


def icon_path(icon_basename: str):
    return resource_path('gui', 'icons', icon_basename)


def internal_plugin_icon_path(plugin_name, icon_basename: str):
    return resource_path('plugins', plugin_name, icon_basename)


@lru_cache(maxsize=1000)
def read_QIcon(icon_basename: str) -> QIcon:
    return QIcon(icon_path(icon_basename))


def read_QPixmap_from_bytes(b: bytes) -> QPixmap:
    qp = QPixmap()
    qp.loadFromData(b)
    return qp


def read_QIcon_from_bytes(b: bytes) -> QIcon:
    qp = read_QPixmap_from_bytes(b)
    return QIcon(qp)


class IconLabel(QWidget):
    HorizontalSpacing = 2
    def __init__(self, *, text='', final_stretch=True, reverse=False, hide_if_empty=False):
        super(QWidget, self).__init__()
        self.hide_if_empty = hide_if_empty
        size = max(16, font_height())
        self.icon_size = QSize(size, size)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.icon = QLabel()
        self.label = QLabel(text)
        self.label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        layout.addWidget(self.icon if reverse else self.label)
        layout.addSpacing(self.HorizontalSpacing)
        layout.addWidget(self.label if reverse else self.icon)
        if final_stretch:
            layout.addStretch()
        self.setText(text)

    def setText(self, text):
        self.label.setText(text)
        if self.hide_if_empty:
            self.setVisible(bool(text))

    def setIcon(self, icon):
        self.icon.setPixmap(icon.pixmap(self.icon_size))
        self.icon.repaint()  # macOS hack for #6269


def char_width_in_lineedit() -> int:
    char_width = QFontMetrics(QLineEdit().font()).averageCharWidth()
    # 'averageCharWidth' seems to underestimate on Windows, hence 'max()'
    return max(9, char_width)


def font_height(widget: QWidget = None) -> int:
    if widget is None:
        widget = QLabel()
    return QFontMetrics(widget.font()).height()


def webopen(url: str):
    if sys.platform == 'linux' and os.environ.get('APPIMAGE'):
        # When on Linux webbrowser.open can fail in AppImage because it can't find the correct libdbus.
        # We just fork the process and unset LD_LIBRARY_PATH before opening the URL.
        # See #5425
        if os.fork() == 0:
            del os.environ['LD_LIBRARY_PATH']
            webbrowser.open(url)
            os._exit(0)
    else:
        webbrowser.open(url)


class FixedAspectRatioLayout(QLayout):
    def __init__(self, parent: QWidget = None, aspect_ratio: float = 1.0):
        super().__init__(parent)
        self.aspect_ratio = aspect_ratio
        self.items: List[QLayoutItem] = []

    def set_aspect_ratio(self, aspect_ratio: float = 1.0):
        self.aspect_ratio = aspect_ratio
        self.update()

    def addItem(self, item: QLayoutItem):
        self.items.append(item)

    def count(self) -> int:
        return len(self.items)

    def itemAt(self, index: int) -> QLayoutItem:
        if index >= len(self.items):
            return None
        return self.items[index]

    def takeAt(self, index: int) -> QLayoutItem:
        if index >= len(self.items):
            return None
        return self.items.pop(index)

    def _get_contents_margins_size(self) -> QSize:
        margins = self.contentsMargins()
        return QSize(margins.left() + margins.right(), margins.top() + margins.bottom())

    def setGeometry(self, rect: QRect):
        super().setGeometry(rect)
        if not self.items:
            return

        contents = self.contentsRect()
        if contents.height() > 0:
            c_aratio = contents.width() / contents.height()
        else:
            c_aratio = 1
        s_aratio = self.aspect_ratio
        item_rect = QRect(QPoint(0, 0), QSize(
            contents.width() if c_aratio < s_aratio else int(contents.height() * s_aratio),
            contents.height() if c_aratio > s_aratio else int(contents.width() / s_aratio)
        ))

        content_margins = self.contentsMargins()
        free_space = contents.size() - item_rect.size()

        for item in self.items:
            if free_space.width() > 0 and not item.alignment() & Qt.AlignmentFlag.AlignLeft:
                if item.alignment() & Qt.AlignmentFlag.AlignRight:
                    item_rect.moveRight(contents.width() + content_margins.right())
                else:
                    item_rect.moveLeft(content_margins.left() + (free_space.width() // 2))
            else:
                item_rect.moveLeft(content_margins.left())

            if free_space.height() > 0 and not item.alignment() & Qt.AlignmentFlag.AlignTop:
                if item.alignment() & Qt.AlignmentFlag.AlignBottom:
                    item_rect.moveBottom(contents.height() + content_margins.bottom())
                else:
                    item_rect.moveTop(content_margins.top() + (free_space.height() // 2))
            else:
                item_rect.moveTop(content_margins.top())

            item.widget().setGeometry(item_rect)

    def sizeHint(self) -> QSize:
        result = QSize()
        for item in self.items:
            result = result.expandedTo(item.sizeHint())
        return self._get_contents_margins_size() + result

    def minimumSize(self) -> QSize:
        result = QSize()
        for item in self.items:
            result = result.expandedTo(item.minimumSize())
        return self._get_contents_margins_size() + result

    def expandingDirections(self) -> Qt.Orientation:
        return Qt.Orientation.Horizontal | Qt.Orientation.Vertical


def QColorLerp(a: QColor, b: QColor, t: float):
    """
    Blends two QColors. t=0 returns a. t=1 returns b. t=0.5 returns evenly mixed.
    """
    t = max(min(t, 1.0), 0.0)
    i_t = 1.0 - t
    return QColor(
        int((a.red()   * i_t) + (b.red()   * t)),
        int((a.green() * i_t) + (b.green() * t)),
        int((a.blue()  * i_t) + (b.blue()  * t)),
        int((a.alpha() * i_t) + (b.alpha() * t)),
    )


class ImageGraphicsEffect(QObject):
    """
    Applies a QGraphicsEffect to a QImage
    """

    def __init__(self, parent: QObject, effect: QGraphicsEffect):
        super().__init__(parent)
        assert effect, 'effect must be set'
        self.effect = effect
        self.graphics_scene = QGraphicsScene()
        self.graphics_item = QGraphicsPixmapItem()
        self.graphics_item.setGraphicsEffect(effect)
        self.graphics_scene.addItem(self.graphics_item)

    def apply(self, image: QImage):
        assert image, 'image must be set'
        result = QImage(image.size(), QImage.Format.Format_ARGB32)
        result.fill(Qt.GlobalColor.transparent)
        painter = QPainter(result)
        self.graphics_item.setPixmap(QPixmap.fromImage(image))
        self.graphics_scene.render(painter)
        self.graphics_item.setPixmap(QPixmap())
        return result


def insert_spaces(text: str, every_chars: int) -> str:
    '''Insert spaces at every Nth character to allow for WordWrap'''
    return ' '.join(text[i:i+every_chars] for i in range(0, len(text), every_chars))


def set_windows_os_screenshot_protection_drm_flag(window: QWidget) -> None:
    """
    sets the windows WDA_MONITOR flag on the window so windows prevents capturing
    screenshots and microsoft recall will not be able to record the window
    https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setwindowdisplayaffinity
    """
    if sys.platform not in ('win32', 'windows'):
        return
    try:
        window_id = int(window.winId())
        WDA_MONITOR = 0x01
        ctypes.windll.user32.SetWindowDisplayAffinity(window_id, WDA_MONITOR)
    except Exception:
        _logger.exception(f"failed to set windows screenshot protection flag")


def debug_widget_layouts(gui_element: QObject):
    """Draw red borders around all widgets of given QObject for debugging.
    E.g. add util.debug_widget_layouts(self) at the end of TxEditor.__init__
    """
    assert isinstance(gui_element, QObject) and hasattr(gui_element, 'findChildren')
    def set_border(widget):
        if widget is not None:
            widget.setStyleSheet(widget.styleSheet() + " * { border: 1px solid red; }")

    # Apply to all child widgets recursively
    for widget in gui_element.findChildren(QWidget):
        set_border(widget)


class _ABCQObjectMeta(type(QObject), ABCMeta): pass
class _ABCQWidgetMeta(type(QWidget), ABCMeta): pass
class AbstractQObject(QObject, ABC, metaclass=_ABCQObjectMeta): pass
class AbstractQWidget(QWidget, ABC, metaclass=_ABCQWidgetMeta): pass


if __name__ == "__main__":
    app = QApplication([])
    t = WaitingDialog(None, 'testing ...', lambda: [time.sleep(1)], lambda x: QMessageBox.information(None, 'done', "done"))
    t.start()
    app.exec()



from typing import Optional

from PyQt6.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject, QVariant

from electrum.i18n import _
from electrum.logging import get_logger
from electrum.bitcoin import DummyAddress
from electrum.util import format_time, TxMinedInfo
from electrum.transaction import tx_from_any, Transaction, PartialTransaction
from electrum.network import Network
from electrum.address_synchronizer import TX_HEIGHT_UNCONF_PARENT, TX_HEIGHT_UNCONFIRMED, TX_HEIGHT_FUTURE
from electrum.wallet import TxSighashDanger
from electrum.fee_policy import FeePolicy

from electrum.gui.common_qt.util import QtEventListener, qt_event_listener

from .qewallet import QEWallet
from .qetypes import QEAmount


class QETxDetails(QObject, QtEventListener):
    _logger = get_logger(__name__)

    confirmRemoveLocalTx = pyqtSignal([str], arguments=['message'])
    txRemoved = pyqtSignal()
    saveTxError = pyqtSignal([str, str], arguments=['code', 'message'])
    saveTxSuccess = pyqtSignal()

    detailsChanged = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.register_callbacks()
        self.destroyed.connect(lambda: self.on_destroy())

        self._wallet = None  # type: Optional[QEWallet]
        self._txid = ''
        self._rawtx = ''
        self._label = ''

        self._tx = None  # type: Optional[Transaction]

        self._status = ''
        self._amount = QEAmount()
        self._lnamount = QEAmount()
        self._fee = QEAmount()
        self._feerate_str = ''
        self._inputs = []
        self._outputs = []

        self._is_lightning_funding_tx = False
        self._can_bump = False
        self._can_dscancel = False
        self._can_broadcast = False
        self._can_cpfp = False
        self._can_save_as_local = False
        self._can_remove = False
        self._can_sign = False
        self._is_unrelated = False
        self._is_complete = False
        self._is_mined = False
        self._is_rbf_enabled = False
        self._is_removed = False
        self._lock_delay = 0
        self._sighash_danger = TxSighashDanger()

        self._mempool_depth = ''
        self._in_mempool = False

        self._date = ''
        self._timestamp = 0
        self._confirmations = 0
        self._header_hash = ''
        self._short_id = ""

    def on_destroy(self):
        self.unregister_callbacks()

    @qt_event_listener
    def on_event_verified(self, wallet, txid, info):
        if wallet == self._wallet.wallet and txid == self._txid:
            self._logger.debug(f'verified event for our txid {txid}')
            self.update()

    @qt_event_listener
    def on_event_new_transaction(self, wallet, tx):
        if wallet == self._wallet.wallet and tx.txid() == self._txid:
            self._logger.debug(f'new_transaction event for our txid {self._txid}')
            self.update()

    @qt_event_listener
    def on_event_removed_transaction(self, wallet, tx):
        if wallet == self._wallet.wallet and tx.txid() == self._txid:
            self._logger.debug(f'removed my transaction {tx.txid()}')
            self._is_removed = True
            self.update()
            self.txRemoved.emit()

    @qt_event_listener
    def on_event_fee_histogram(self, histogram):
        if not self._wallet or not self._tx:
            return
        self.update()

    walletChanged = pyqtSignal()
    @pyqtProperty(QVariant, notify=walletChanged)
    def wallet(self) -> QEWallet:
        return self._wallet

    @wallet.setter
    def wallet(self, wallet: QEWallet):
        assert wallet is None or isinstance(wallet, QEWallet)
        if self._wallet != wallet:
            self._wallet = wallet
            self.walletChanged.emit()

    txidChanged = pyqtSignal()
    @pyqtProperty(str, notify=txidChanged)
    def txid(self):
        return self._txid

    @txid.setter
    def txid(self, txid: str):
        if self._txid != txid:
            self._logger.debug(f'txid set -> {txid}')
            self._txid = txid
            self.txidChanged.emit()
            self.update(from_txid=True)

    @pyqtProperty(str, notify=detailsChanged)
    def rawtx(self):
        return self._rawtx

    @rawtx.setter
    def rawtx(self, rawtx: str):
        if self._rawtx != rawtx:
            self._logger.debug(f'rawtx set -> {rawtx}')
            self._rawtx = rawtx
            if not rawtx:
                return
            try:
                self._tx = tx_from_any(rawtx, deserialize=True)
                self._txid = self._tx.txid()
                self.txidChanged.emit()
                self.update()
            except Exception as e:
                self._tx = None
                self._logger.error(repr(e))

    labelChanged = pyqtSignal()
    @pyqtProperty(str, notify=labelChanged)
    def label(self):
        return self._label

    @pyqtSlot(str)
    def setLabel(self, label: str):
        if label != self._label:
            self._wallet.wallet.set_label(self._txid, label)
            self._label = label
            self.labelChanged.emit()

    @pyqtProperty(str, notify=detailsChanged)
    def status(self):
        return self._status

    @pyqtProperty(str, notify=detailsChanged)
    def warning(self):
        return self._sighash_danger.get_long_message()

    @pyqtProperty(QEAmount, notify=detailsChanged)
    def amount(self):
        return self._amount

    @pyqtProperty(QEAmount, notify=detailsChanged)
    def lnAmount(self):
        return self._lnamount

    @pyqtProperty(QEAmount, notify=detailsChanged)
    def fee(self):
        return self._fee

    @pyqtProperty(str, notify=detailsChanged)
    def feeRateStr(self):
        return self._feerate_str

    @pyqtProperty('QVariantList', notify=detailsChanged)
    def inputs(self):
        return self._inputs

    @pyqtProperty('QVariantList', notify=detailsChanged)
    def outputs(self):
        return self._outputs

    @pyqtProperty(bool, notify=detailsChanged)
    def isMined(self):
        return self._is_mined

    @pyqtProperty(bool, notify=detailsChanged)
    def isRemoved(self):
        return self._is_removed

    @pyqtProperty(str, notify=detailsChanged)
    def mempoolDepth(self):
        return self._mempool_depth

    @pyqtProperty(bool, notify=detailsChanged)
    def inMempool(self):
        return self._in_mempool

    @pyqtProperty(str, notify=detailsChanged)
    def date(self):
        return self._date

    @pyqtProperty(int, notify=detailsChanged)
    def timestamp(self):
        return self._timestamp

    @pyqtProperty(int, notify=detailsChanged)
    def confirmations(self):
        return self._confirmations

    @pyqtProperty(str, notify=detailsChanged)
    def shortId(self):
        return self._short_id

    @pyqtProperty(str, notify=detailsChanged)
    def headerHash(self):
        return self._header_hash

    @pyqtProperty(bool, notify=detailsChanged)
    def isLightningFundingTx(self):
        return self._is_lightning_funding_tx

    @pyqtProperty(bool, notify=detailsChanged)
    def canBump(self):
        return self._can_bump

    @pyqtProperty(bool, notify=detailsChanged)
    def canCancel(self):
        return self._can_dscancel

    @pyqtProperty(bool, notify=detailsChanged)
    def canBroadcast(self):
        return self._can_broadcast

    @pyqtProperty(bool, notify=detailsChanged)
    def canCpfp(self):
        return self._can_cpfp

    @pyqtProperty(bool, notify=detailsChanged)
    def canSaveAsLocal(self):
        return self._can_save_as_local

    @pyqtProperty(bool, notify=detailsChanged)
    def canRemove(self):
        return self._can_remove

    @pyqtProperty(bool, notify=detailsChanged)
    def canSign(self):
        return self._can_sign

    @pyqtProperty(bool, notify=detailsChanged)
    def isUnrelated(self):
        return self._is_unrelated

    @pyqtProperty(bool, notify=detailsChanged)
    def isComplete(self):
        return self._is_complete

    @pyqtProperty(bool, notify=detailsChanged)
    def isRbfEnabled(self):
        return self._is_rbf_enabled

    @pyqtProperty(int, notify=detailsChanged)
    def lockDelay(self):
        return self._lock_delay

    @pyqtProperty(bool, notify=detailsChanged)
    def shouldConfirm(self):
        return self._sighash_danger.needs_confirm()

    def update(self, from_txid: bool = False):
        assert self._wallet

        if self._is_removed:
            self._logger.debug('tx removed, disable gui options')
            self._can_broadcast = False
            self._can_bump = False
            self._can_dscancel = False
            self._can_cpfp = False
            self._can_save_as_local = False
            self._can_remove = False
            self._can_sign = False
            self._mempool_depth = ''
            self._status = _('removed')
            self.detailsChanged.emit()
            return

        if from_txid:
            self._tx = self._wallet.wallet.db.get_transaction(self._txid)
            assert self._tx is not None, f'unknown txid "{self._txid}"'

        #self._logger.debug(repr(self._tx.to_json()))

        self._logger.debug('adding info from wallet')
        self._tx.add_info_from_wallet(self._wallet.wallet)
        if not self._tx.is_complete() and self._tx.is_missing_info_from_network():
            Network.run_from_another_thread(
                self._tx.add_info_from_network(self._wallet.wallet.network, timeout=10))  # FIXME is this needed?...

        sm = self._wallet.wallet.lnworker.swap_manager if self._wallet.wallet.lnworker else None

        self._inputs = list(map(lambda x: {
            'short_id': x.prevout.short_name(),
            'value': x.value_sats(),
            'address': x.address,
            'is_mine': self._wallet.wallet.is_mine(x.address),
            'is_change': self._wallet.wallet.is_change(x.address),
            'is_swap': False if not sm else sm.is_lockup_address_for_a_swap(x.address) or x.address == DummyAddress.SWAP,
            'is_accounting': self._wallet.wallet.is_accounting_address(x.address)
        }, self._tx.inputs()))
        self._outputs = list(map(lambda x: {
            'address': x.get_ui_address_str(),
            'value': QEAmount(amount_sat=x.value),
            'short_id': '',  # TODO
            'is_mine': self._wallet.wallet.is_mine(x.get_ui_address_str()),
            'is_change': self._wallet.wallet.is_change(x.get_ui_address_str()),
            'is_billing': self._wallet.wallet.is_billing_address(x.get_ui_address_str()),
            'is_swap': False if not sm else sm.is_lockup_address_for_a_swap(x.get_ui_address_str()) or x.get_ui_address_str() == DummyAddress.SWAP,
            'is_accounting': self._wallet.wallet.is_accounting_address(x.get_ui_address_str())
        }, self._tx.outputs()))

        txinfo = self._wallet.wallet.get_tx_info(self._tx)

        self._logger.debug(repr(txinfo))

        # can be None if outputs unrelated to wallet seed,
        # e.g. to_local local_force_close commitment CSV-locked p2wsh script
        if txinfo.amount is None:
            self._amount.satsInt = 0
        else:
            self._amount.satsInt = txinfo.amount

        self._status = txinfo.status
        self._fee.satsInt = txinfo.fee

        self._feerate_str = ""
        if txinfo.fee is not None:
            size = self._tx.estimated_size()
            fee_per_kb = txinfo.fee / size * 1000
            self._feerate_str = self._wallet.wallet.config.format_fee_rate(fee_per_kb)

        self._sighash_danger = TxSighashDanger()

        self._lock_delay = 0
        self._in_mempool = False
        self._is_mined = False if not txinfo.tx_mined_status else txinfo.tx_mined_status.height() > 0
        if self._is_mined:
            self.update_mined_status(txinfo.tx_mined_status)
        else:
            if txinfo.tx_mined_status.height() in [TX_HEIGHT_UNCONFIRMED, TX_HEIGHT_UNCONF_PARENT]:
                self._mempool_depth = FeePolicy.depth_tooltip(txinfo.mempool_depth_bytes)
                self._in_mempool = True
            elif txinfo.tx_mined_status.height() == TX_HEIGHT_FUTURE:
                self._lock_delay = txinfo.tx_mined_status.wanted_height - self._wallet.wallet.adb.get_local_height()
            if isinstance(self._tx, PartialTransaction):
                self._sighash_danger = self._wallet.wallet.check_sighash(self._tx)

        if self._wallet.wallet.lnworker:
            # Calling wallet.get_full_history here is inefficient.
            # We should probably pass the tx_item to the constructor.
            full_history = self._wallet.wallet.get_full_history()
            item = full_history.get('group:' + self._txid)
            self._lnamount.satsInt = int(item['ln_value'].value) if item else 0
        else:
            self._lnamount.satsInt = 0

        self._is_complete = self._tx.is_complete()
        self._is_rbf_enabled = self._tx.is_rbf_enabled()
        self._is_unrelated = txinfo.amount is None and self._lnamount.isEmpty
        self._is_lightning_funding_tx = txinfo.is_lightning_funding_tx
        self._can_broadcast = txinfo.can_broadcast
        self._can_bump = txinfo.can_bump
        self._can_dscancel = txinfo.can_dscancel
        self._can_cpfp = txinfo.can_cpfp
        self._can_save_as_local = txinfo.can_save_as_local
        self._can_remove = txinfo.can_remove
        self._can_sign = (
            not self._is_complete
            and self._wallet.wallet.can_sign(self._tx)
            and not self._sighash_danger.needs_reject()
        )

        self.detailsChanged.emit()

        if self._txid:
            label = self._wallet.wallet.get_label_for_txid(self._txid)
            if self._label != label:
                self._label = label
                self.labelChanged.emit()

    def update_mined_status(self, tx_mined_info: TxMinedInfo):
        self._mempool_depth = ''
        self._date = format_time(tx_mined_info.timestamp)
        self._timestamp = tx_mined_info.timestamp
        self._confirmations = tx_mined_info.conf
        self._header_hash = tx_mined_info.header_hash
        self._short_id = tx_mined_info.short_id() or ""

    @pyqtSlot()
    def signAndBroadcast(self):
        self._sign(broadcast=True)

    @pyqtSlot()
    def sign(self):
        self._sign(broadcast=False)

    def _sign(self, broadcast):
        # TODO: connecting/disconnecting signal handlers here is hmm
        try:
            if broadcast:
                self._wallet.broadcastSucceeded.disconnect(self.onBroadcastSucceeded)
                self._wallet.broadcastFailed.disconnect(self.onBroadcastFailed)
        except Exception:
            pass

        if broadcast:
            self._wallet.broadcastSucceeded.connect(self.onBroadcastSucceeded)
            self._wallet.broadcastFailed.connect(self.onBroadcastFailed)
            self._wallet.sign_and_broadcast(self._tx, on_success=self.on_signed_tx)
        else:
            self._wallet.sign(self._tx, on_success=self.on_signed_tx)

        # side-effect: signing updates self._tx
        # we rely on this for broadcast

    def on_signed_tx(self, tx: Transaction):
        self._logger.debug('on_signed_tx')
        self.update()

    @pyqtSlot()
    def broadcast(self):
        assert self._tx.is_complete()

        try:
            self._wallet.broadcastFailed.disconnect(self.onBroadcastFailed)
        except Exception:
            pass
        self._wallet.broadcastFailed.connect(self.onBroadcastFailed)

        self._can_broadcast = False
        self.detailsChanged.emit()

        self._wallet.broadcast(self._tx)

    @pyqtSlot(str)
    def onBroadcastSucceeded(self, txid):
        if txid != self._txid:
            return

        self._logger.debug('onBroadcastSucceeded')
        try:
            self._wallet.broadcastSucceeded.disconnect(self.onBroadcastSucceeded)
        except Exception:
            pass

        self._can_broadcast = False
        self.detailsChanged.emit()

    @pyqtSlot(str, str, str)
    def onBroadcastFailed(self, txid, code, reason):
        if txid != self._txid:
            return

        try:
            self._wallet.broadcastFailed.disconnect(self.onBroadcastFailed)
        except Exception:
            pass

        self._can_broadcast = True
        self.detailsChanged.emit()

    @pyqtSlot()
    @pyqtSlot(bool)
    def removeLocalTx(self, confirm=False):
        assert self._can_remove, 'cannot remove'
        txid = self._txid
        assert txid, 'txid unset'

        if not confirm:
            num_child_txs = len(self._wallet.wallet.adb.get_depending_transactions(txid))
            question = _("Are you sure you want to remove this transaction?")
            if num_child_txs > 0:
                question = (
                    _("Are you sure you want to remove this transaction and {} child transactions?")
                    .format(num_child_txs))
            self.confirmRemoveLocalTx.emit(question)
            return

        self._wallet.wallet.adb.remove_transaction(txid)
        self._wallet.wallet.save_db()

        # NOTE: from here, the tx/txid is unknown and all properties are invalid.
        # UI should close TxDetails and avoid interacting with this qetxdetails instance.
        self._tx = None

    @pyqtSlot()
    def save(self):
        if not self._tx:
            return

        if self._wallet.save_tx(self._tx):
            self._can_save_as_local = False
            self._can_remove = True
            self.detailsChanged.emit()

    @pyqtSlot(result='QVariantList')
    def getSerializedTx(self):
        txqr = self._tx.to_qr_data()
        label = ""
        if txid := self._tx.txid():
            label = self._wallet.wallet.get_label_for_txid(txid)
        return [str(self._tx), txqr[0], txqr[1], label]



#!/usr/bin/env python3
"""
Unit tests for vandalism detection using the real LLM API.
Tests known vandalism cases from electrum-locale repository.

Run LLM tests with API environment variables, e.g.:
OPENAI_BASE_URL=https://api.ppq.ai OPENAI_MODEL=google/gemini-3-flash-preview OPENAI_API_KEY=ABC

Offline tests only (diff parsing):
python3 -m unittest -k DiffPars -k UnescapePo -k TestExtract test_llm_proofreader
"""

import asyncio
import os
import textwrap
import unittest

import aiohttp

from llm_proofreader import (
    classify_translation_async, get_concurrency, parse_po_diff, _unescape_po, _extract_po_string_lines,
    _extract_pairs_from_lines,
)


class TestDiffParsingBasic(unittest.TestCase):
    """Offline tests for parse_po_diff — no API key required."""

    def test_empty_diff(self):
        result = parse_po_diff("")
        self.assertEqual(result, [])

    def test_diff_no_po_files(self):
        diff = textwrap.dedent("""\
        diff --git a/README.md b/README.md
        --- a/README.md
        +++ b/README.md
        @@ -1,2 +1,2 @@
        -old line
        +new line
         context
        """)
        result = parse_po_diff(diff)
        self.assertEqual(result, [])

    def test_new_translation_added(self):
        """A previously empty msgstr gets a translation."""
        diff = textwrap.dedent("""\
        diff --git a/locale/de_DE/electrum.po b/locale/de_DE/electrum.po
        --- a/locale/de_DE/electrum.po
        +++ b/locale/de_DE/electrum.po
        @@ -10,3 +10,3 @@
         #: electrum/gui/qt/main_window.py:100
         msgid "Wallet"
        -msgstr ""
        +msgstr "Geldbeutel"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 1)
        locale, msgid, msgstr = result[0]
        self.assertEqual(locale, "de_DE")
        self.assertEqual(msgid, "Wallet")
        self.assertEqual(msgstr, "Geldbeutel")

    def test_translation_modified(self):
        """An existing translation is changed."""
        diff = textwrap.dedent("""\
        diff --git a/locale/fr_FR/electrum.po b/locale/fr_FR/electrum.po
        --- a/locale/fr_FR/electrum.po
        +++ b/locale/fr_FR/electrum.po
        @@ -5,3 +5,3 @@
         #: electrum/gui/qt/main_window.py:200
         msgid "Send"
        -msgstr "Envoyer"
        +msgstr "Transmettre"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], ("fr_FR", "Send", "Transmettre"))

    def test_unchanged_translation_ignored(self):
        """Context-only changes (line number updates) should not be flagged."""
        diff = textwrap.dedent("""\
        diff --git a/locale/de_DE/electrum.po b/locale/de_DE/electrum.po
        --- a/locale/de_DE/electrum.po
        +++ b/locale/de_DE/electrum.po
        @@ -10,3 +10,3 @@
        -#: electrum/gui/qt/main_window.py:100
        +#: electrum/gui/qt/main_window.py:102
         msgid "Wallet"
         msgstr "Geldbeutel"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(result, [])

    def test_empty_msgstr_ignored(self):
        """Newly added entries with empty msgstr should not be returned."""
        diff = textwrap.dedent("""\
        diff --git a/locale/ja_JP/electrum.po b/locale/ja_JP/electrum.po
        --- a/locale/ja_JP/electrum.po
        +++ b/locale/ja_JP/electrum.po
        @@ -100,3 +100,6 @@
         msgid "Old entry"
         msgstr "古いエントリ"
         
        +msgid "Brand new entry"
        +msgstr ""
        +
        """)
        result = parse_po_diff(diff)
        self.assertEqual(result, [])

    def test_multiple_locales(self):
        """Changes across multiple locale files are all captured."""
        diff = textwrap.dedent("""\
        diff --git a/locale/de_DE/electrum.po b/locale/de_DE/electrum.po
        --- a/locale/de_DE/electrum.po
        +++ b/locale/de_DE/electrum.po
        @@ -10,2 +10,2 @@
         msgid "Send"
        -msgstr ""
        +msgstr "Senden"
         
        diff --git a/locale/fr_FR/electrum.po b/locale/fr_FR/electrum.po
        --- a/locale/fr_FR/electrum.po
        +++ b/locale/fr_FR/electrum.po
        @@ -10,2 +10,2 @@
         msgid "Send"
        -msgstr ""
        +msgstr "Envoyer"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 2)
        locales = {r[0] for r in result}
        self.assertEqual(locales, {"de_DE", "fr_FR"})

    def test_multiline_msgstr(self):
        """Multi-line msgstr values are correctly joined."""
        diff = textwrap.dedent("""\
        diff --git a/locale/de_DE/electrum.po b/locale/de_DE/electrum.po
        --- a/locale/de_DE/electrum.po
        +++ b/locale/de_DE/electrum.po
        @@ -10,2 +10,4 @@
         msgid "Long text"
        -msgstr ""
        +msgstr "Erste Zeile "
        +"Zweite Zeile "
        +"Dritte Zeile"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0][2], "Erste Zeile Zweite Zeile Dritte Zeile")

    def test_multiline_msgid_and_msgstr(self):
        """Multi-line msgid and msgstr are both correctly parsed."""
        diff = textwrap.dedent("""\
        diff --git a/locale/es_ES/electrum.po b/locale/es_ES/electrum.po
        --- a/locale/es_ES/electrum.po
        +++ b/locale/es_ES/electrum.po
        @@ -10,4 +10,4 @@
         msgid ""
         "First part "
         "second part"
        -msgstr ""
        +msgstr "Primera parte segunda parte"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0][1], "First part second part")
        self.assertEqual(result[0][2], "Primera parte segunda parte")

    def test_spam_injection_detected_in_diff(self):
        """A vandalized translation in a diff is correctly extracted for checking."""
        diff = textwrap.dedent("""\
        diff --git a/locale/fr_FR/electrum.po b/locale/fr_FR/electrum.po
        --- a/locale/fr_FR/electrum.po
        +++ b/locale/fr_FR/electrum.po
        @@ -10,2 +10,2 @@
         msgid "Avoid spending from used addresses"
        -msgstr ""
        +msgstr "bc1qgdl5a90ccznwteha436fn52nekdwuu9ld32n3c"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0][0], "fr_FR")
        self.assertEqual(result[0][1], "Avoid spending from used addresses")
        self.assertEqual(result[0][2], "bc1qgdl5a90ccznwteha436fn52nekdwuu9ld32n3c")

    def test_header_entry_skipped(self):
        """The header entry (empty msgid) should be skipped."""
        diff = textwrap.dedent("""\
        diff --git a/locale/de_DE/electrum.po b/locale/de_DE/electrum.po
        --- a/locale/de_DE/electrum.po
        +++ b/locale/de_DE/electrum.po
        @@ -2,4 +2,4 @@
         msgid ""
         msgstr ""
        -"POT-Creation-Date: 2025-08-23 15:01+0000\\n"
        -"PO-Revision-Date: 2025-08-23 15:01\\n"
        +"POT-Creation-Date: 2026-01-22 09:38+0000\\n"
        +"PO-Revision-Date: 2026-01-22 09:38\\n"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(result, [])

    def test_deleted_translation_ignored(self):
        """A translation that was removed (msgstr goes from non-empty to empty) should not be flagged."""
        diff = textwrap.dedent("""\
        diff --git a/locale/de_DE/electrum.po b/locale/de_DE/electrum.po
        --- a/locale/de_DE/electrum.po
        +++ b/locale/de_DE/electrum.po
        @@ -10,2 +10,2 @@
         msgid "Wallet"
        -msgstr "Geldbeutel"
        +msgstr ""
        """)
        result = parse_po_diff(diff)
        self.assertEqual(result, [])

    def test_multiple_hunks_same_file(self):
        """Multiple hunks in the same file are all processed."""
        diff = textwrap.dedent("""\
        diff --git a/locale/it_IT/electrum.po b/locale/it_IT/electrum.po
        --- a/locale/it_IT/electrum.po
        +++ b/locale/it_IT/electrum.po
        @@ -10,2 +10,2 @@
         msgid "Send"
        -msgstr ""
        +msgstr "Invia"
         
        @@ -50,2 +50,2 @@
         msgid "Receive"
        -msgstr ""
        +msgstr "Ricevi"
        """)
        result = parse_po_diff(diff)
        self.assertEqual(len(result), 2)
        msgids = {r[1] for r in result}
        self.assertEqual(msgids, {"Send", "Receive"})


class TestDiffParsingTypicalDiff(unittest.TestCase):
    """Tests using the actual unittest_diff.diff file (offline, no API)."""

    @classmethod
    def setUpClass(cls):
        diff_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "unittest_diff_974d671_eab55b5.diff",
        )
        with open(diff_path, "r", encoding="utf-8", errors="replace") as f:
            cls.diff_text = f.read()
        cls.parsed_diff = parse_po_diff(cls.diff_text)

    def test_diff_returns_results(self):
        """The diff should contain 560 changed translations."""
        self.assertEqual(len(self.parsed_diff), 560, "Expected 560 changed translations")

    def test_diff_all_have_locale(self):
        """Every result should have a non-empty locale."""
        for locale, msgid, msgstr in self.parsed_diff:
            self.assertTrue(locale, f"Empty locale for msgid={msgid!r}")

    def test_diff_no_empty_msgstr(self):
        """No result should have an empty msgstr."""
        for locale, msgid, msgstr in self.parsed_diff:
            self.assertTrue(msgstr, f"Empty msgstr for {locale}:{msgid!r}")

    def test_diff_no_empty_msgid(self):
        """No result should have an empty msgid (header entries should be skipped)."""
        for locale, msgid, msgstr in self.parsed_diff:
            self.assertTrue(msgid, f"Empty msgid for {locale}:{msgstr!r}")

    def test_diff_multiple_locales(self):
        """The diff should contain changes from 20 locales."""
        locales = {r[0] for r in self.parsed_diff}
        self.assertEqual(len(locales), 20, "Expected 20 locales in diff")


class TestUnescapePo(unittest.TestCase):
    """Tests for _unescape_po helper."""

    def test_newline(self):
        self.assertEqual(_unescape_po("hello\\nworld"), "hello\nworld")

    def test_tab(self):
        self.assertEqual(_unescape_po("hello\\tworld"), "hello\tworld")

    def test_quote(self):
        self.assertEqual(_unescape_po('say \\"hello\\"'), 'say "hello"')

    def test_backslash(self):
        self.assertEqual(_unescape_po("path\\\\file"), "path\\file")

    def test_no_escapes(self):
        self.assertEqual(_unescape_po("plain text"), "plain text")


class TestExtractPoStringLines(unittest.TestCase):
    """Tests for _extract_po_string_lines helper."""

    def test_single_line(self):
        lines = ['msgid "Hello"', 'msgstr "World"']
        val, idx = _extract_po_string_lines(lines, 0)
        self.assertEqual(val, "Hello")
        self.assertEqual(idx, 1)

    def test_multi_line(self):
        lines = ['msgstr ""', '"First "', '"Second"', '']
        val, idx = _extract_po_string_lines(lines, 0)
        self.assertEqual(val, "First Second")
        self.assertEqual(idx, 3)

    def test_continuation_stops_at_non_quoted(self):
        lines = ['msgid "Test"', '#: comment', 'msgstr "Value"']
        val, idx = _extract_po_string_lines(lines, 0)
        self.assertEqual(val, "Test")
        self.assertEqual(idx, 1)


class TestExtractPairsFromLines(unittest.TestCase):
    """Tests for _extract_pairs_from_lines helper."""

    def test_simple_pair(self):
        lines = ['msgid "Hello"', 'msgstr "Hola"']
        pairs = _extract_pairs_from_lines(lines)
        self.assertEqual(pairs, [("Hello", "Hola")])

    def test_multiple_pairs(self):
        lines = [
            'msgid "One"', 'msgstr "Uno"', '',
            'msgid "Two"', 'msgstr "Dos"',
        ]
        pairs = _extract_pairs_from_lines(lines)
        self.assertEqual(len(pairs), 2)

    def test_empty_msgid_skipped(self):
        lines = ['msgid ""', 'msgstr "Header content"', '', 'msgid "Real"', 'msgstr "Actual"']
        pairs = _extract_pairs_from_lines(lines)
        self.assertEqual(len(pairs), 1)
        self.assertEqual(pairs[0][0], "Real")


class AsyncVandalismTestCase(unittest.IsolatedAsyncioTestCase):
    """Base class that provides a shared aiohttp session and semaphore."""

    _session: aiohttp.ClientSession = None
    _semaphore: asyncio.Semaphore = None

    async def asyncSetUp(self):
        self._session = aiohttp.ClientSession()
        self._semaphore = asyncio.Semaphore(get_concurrency())

    async def asyncTearDown(self):
        await self._session.close()

    async def classify(self, msgid: str, msgstr: str, lang: str) -> str:
        return await classify_translation_async(
            self._session, self._semaphore, msgid, msgstr, lang
        )


class TestVandalismDetection(AsyncVandalismTestCase):
    """
    Tests that known vandalism cases are correctly classified as Spam.
    These are real examples from the electrum-locale repository.
    """

    async def test_fr_FR_redo_thongchai(self):
        """fr_FR: '&Redo' translated as 'Thongchai' (unrelated name)"""
        result = await self.classify("&Redo", "Thongchai", "fr_FR")
        self.assertEqual(result, "Spam", "Unrelated name substitution should be Spam")

    async def test_fr_FR_cancellation_txid(self):
        """fr_FR: 'Cancellation Transaction ID' translated as actual txid hash"""
        result = await self.classify(
            "Cancellation Transaction ID",
            "edbd3224d120d8304d2194287c4b267707b50b2be46a78409d701327d9d47476\n",
            "fr_FR"
        )
        self.assertEqual(result, "Spam", "Injected transaction hash should be Spam")

    async def test_fa_IR_script_shah2026(self):
        """fa_IR: '- an arbitrary on-chain script, e.g.:' translated as 'shah2026'"""
        result = await self.classify(
            "- an arbitrary on-chain script, e.g.:",
            "shah2026",
            "fa_IR"
        )
        self.assertEqual(result, "Spam", "Username/handle substitution should be Spam")

    async def test_ro_RO_bitcoin_address_romana(self):
        """ro_RO: '- a Bitcoin address or BIP21 URI' translated as 'Română română'"""
        result = await self.classify(
            "- a Bitcoin address or BIP21 URI",
            "Română română ",
            "ro_RO"
        )
        self.assertEqual(result, "Spam", "Meaningless language name substitution should be Spam")

    async def test_vi_VN_amount_number(self):
        """vi_VN: 'Amount to be sent' translated as '10000000000' (8 years old)"""
        result = await self.classify(
            "Amount to be sent",
            "10000000000",
            "vi_VN"
        )
        self.assertEqual(result, "Spam", "Random number substitution should be Spam")

    async def test_cs_CZ_paper_email(self):
        """cs_CZ: Instructions translated as email address (2 years old)"""
        result = await self.classify(
            "1. Place this paper on a flat and well illuminated surface.",
            "veronikatilke33@seznam.cz",
            "cs_CZ"
        )
        self.assertEqual(result, "Spam", "Email address injection should be Spam")

    async def test_es_ES_amount_received_email(self):
        """es_ES: 'Amount received: {}' translated as email address"""
        result = await self.classify(
            "Amount received: {}",
            "cmc6686@gmail.com",
            "es_ES"
        )
        self.assertEqual(result, "Spam", "Email address injection should be Spam")

    async def test_fa_IR_undo_username(self):
        """fa_IR: '&Undo' translated as 'Rastakh8888'"""
        result = await self.classify(
            "&Undo",
            "Rastakh8888\n",
            "fa_IR"
        )
        self.assertEqual(result, "Spam", "Username substitution should be Spam")

    async def test_fr_FR_avoid_spending_btc_address(self):
        """fr_FR: Address usage warning translated as BTC address (stealing-oriented)"""
        result = await self.classify(
            "Avoid spending from used addresses",
            "bc1qgdl5a90ccznwteha436fn52nekdwuu9ld32n3c",
            "fr_FR"
        )
        self.assertEqual(result, "Spam", "Bitcoin address injection should be Spam")

    async def test_ar_SA_alert_address_btc_address(self):
        """ar_SA: 'Alert Address' translated as BTC address (stealing-oriented)"""
        result = await self.classify(
            "Alert Address",
            "bc1q393et0s9tv82udvjqfrupyv90s3fqvcfglhrea",
            "ar_SA"
        )
        self.assertEqual(result, "Spam", "Bitcoin address injection should be Spam")


class TestVandalismURL(AsyncVandalismTestCase):
    """Tests for URL injection vandalism."""

    async def test_url_injection_random(self):
        """URL injected into unrelated translation"""
        result = await self.classify(
            "Wallet settings",
            "https://freebitcoin.com/claim-now",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "Random URL injection should be Spam")

    async def test_url_injection_electrum_update(self):
        """URL injected into unrelated translation"""
        result = await self.classify(
            "Download the update from the Electrum website.",
            "Lade die Electrum Aktualisierung von electrum.io herunter.",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "Electrum update URL injection should be Spam")

    async def test_telegram_link_injection(self):
        """Telegram link injected"""
        result = await self.classify(
            "Please wait...",
            "t.me/cryptotrader_group",
            "ru_RU"
        )
        self.assertEqual(result, "Spam", "Telegram link injection should be Spam")

    async def test_phone_number_injection(self):
        """Phone number injected instead of translation"""
        result = await self.classify(
            "Network status",
            "+1-800-555-0199",
            "es_ES"
        )
        self.assertEqual(result, "Spam", "Phone number injection should be Spam")


class TestVandalismPoliticalAdvertising(AsyncVandalismTestCase):
    """Tests for political/advertising vandalism."""

    async def test_advertising_injection(self):
        """Advertising text injected"""
        result = await self.classify(
            "Confirm transaction",
            "Buy cheap electronics at bestdeals.com!",
            "it_IT"
        )
        self.assertEqual(result, "Spam", "Advertising injection should be Spam")

    async def test_political_slogan(self):
        """Political slogan injected"""
        result = await self.classify(
            "Balance",
            "Vote for candidate XYZ 2026",
            "pt_BR"
        )
        self.assertEqual(result, "Spam", "Political slogan injection should be Spam")


class TestVandalismGibberish(AsyncVandalismTestCase):
    """Tests for gibberish/random content."""

    async def test_random_characters(self):
        """Random character sequence"""
        result = await self.classify(
            "Transaction fee",
            "asdfghjkl qwerty zxcvbnm",
            "nl_NL"
        )
        self.assertEqual(result, "Spam", "Random characters should be Spam")

    async def test_keyboard_mash(self):
        """Keyboard mashing"""
        result = await self.classify(
            "Enter password",
            "jjjjjjjjjjjjjjjjjjjj",
            "pl_PL"
        )
        self.assertEqual(result, "Spam", "Keyboard mash should be Spam")

    async def test_emoji_spam(self):
        """Emoji spam instead of translation"""
        result = await self.classify(
            "Send Bitcoin",
            "\U0001f680\U0001f680\U0001f680\U0001f4b0\U0001f4b0\U0001f4b0\U0001f525\U0001f525\U0001f525",
            "ko_KR"
        )
        self.assertEqual(result, "Spam", "Emoji spam should be Spam")


class TestVandalismCryptoAddresses(AsyncVandalismTestCase):
    """Tests for cryptocurrency address injection."""

    async def test_ethereum_address_injection(self):
        """Ethereum address injected"""
        result = await self.classify(
            "Receive",
            "0x742d35Cc6634C0532925a3b844Bc9e7595f8eA12",
            "zh_CN"
        )
        self.assertEqual(result, "Spam", "Ethereum address injection should be Spam")

    async def test_litecoin_address_injection(self):
        """Litecoin address injected"""
        result = await self.classify(
            "Copy address",
            "LTC1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t4",
            "tr_TR"
        )
        self.assertEqual(result, "Spam", "Litecoin address injection should be Spam")

    async def test_monero_address_injection(self):
        """Monero address injected"""
        result = await self.classify(
            "Wallet backup",
            "44AFFq5kSiGBoZ4NMDwYtN18obc8AemS33DBLWs3H7otXft3XjrpDtQGv7SqSsaBYBb98uNbr2VBBEt7f2wfn3RVGQBEP3A",
            "sv_SE"
        )
        self.assertEqual(result, "Spam", "Monero address injection should be Spam")


class TestVandalismSocialMedia(AsyncVandalismTestCase):
    """Tests for social media handle injection."""

    async def test_twitter_handle_injection(self):
        """Twitter/X handle injected"""
        result = await self.classify(
            "Help",
            "@follow_me_for_free_btc",
            "hu_HU"
        )
        self.assertEqual(result, "Spam", "Twitter handle injection should be Spam")

    async def test_discord_injection(self):
        """Discord server link injected"""
        result = await self.classify(
            "About",
            "discord.gg/freecrypto",
            "cs_CZ"
        )
        self.assertEqual(result, "Spam", "Discord link injection should be Spam")


class TestVandalismWrongLanguage(AsyncVandalismTestCase):
    """Tests for completely wrong target language."""

    async def test_wrong_script_cyrillic_for_japanese(self):
        """Cyrillic text for Japanese target"""
        result = await self.classify(
            "Settings",
            "Привет мир",
            "ja_JP"
        )
        self.assertEqual(result, "Spam", "Cyrillic for Japanese target should be Spam")

    async def test_wrong_script_arabic_for_german(self):
        """Arabic text for German target"""
        result = await self.classify(
            "Confirm",
            "مرحبا بالعالم",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "Arabic for German target should be Spam")


class TestGenuineTranslations(AsyncVandalismTestCase):
    """
    Control tests to ensure legitimate translations are not flagged.
    """

    async def test_genuine_fr_FR(self):
        """French translation should be Genuine"""
        result = await self.classify(
            "Amount to be sent",
            "Montant à envoyer",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Valid French translation should be Genuine")

    async def test_genuine_de_DE(self):
        """German translation should be Genuine"""
        result = await self.classify(
            "Bitcoin Address",
            "Bitcoin-Adresse",
            "de_DE"
        )
        self.assertEqual(result, "Genuine", "Valid German translation should be Genuine")

    async def test_genuine_ja_JP(self):
        """Japanese translation should be Genuine"""
        result = await self.classify(
            "Send",
            "送金",
            "ja_JP"
        )
        self.assertEqual(result, "Genuine", "Valid Japanese translation should be Genuine")

    async def test_genuine_technical_terms_preserved(self):
        """Technical terms left untranslated should be Genuine"""
        result = await self.classify(
            "BIP21 URI",
            "URI BIP21",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Technical terms preservation should be Genuine")

    async def test_genuine_zh_CN(self):
        """Chinese simplified translation should be Genuine"""
        result = await self.classify(
            "Wallet",
            "钱包",
            "zh_CN"
        )
        self.assertEqual(result, "Genuine", "Valid Chinese translation should be Genuine")

    async def test_genuine_ar_SA(self):
        """Arabic translation should be Genuine"""
        result = await self.classify(
            "Password",
            "كلمة المرور",
            "ar_SA"
        )
        self.assertEqual(result, "Genuine", "Valid Arabic translation should be Genuine")

    async def test_genuine_ru_RU(self):
        """Russian translation should be Genuine"""
        result = await self.classify(
            "Transaction",
            "Транзакция",
            "ru_RU"
        )
        self.assertEqual(result, "Genuine", "Valid Russian translation should be Genuine")

    async def test_genuine_ko_KR(self):
        """Korean translation should be Genuine"""
        result = await self.classify(
            "Confirm",
            "확인",
            "ko_KR"
        )
        self.assertEqual(result, "Genuine", "Valid Korean translation should be Genuine")

    async def test_genuine_pt_BR(self):
        """Portuguese (Brazil) translation should be Genuine"""
        result = await self.classify(
            "Receive",
            "Receber",
            "pt_BR"
        )
        self.assertEqual(result, "Genuine", "Valid Portuguese translation should be Genuine")

    async def test_genuine_placeholder_preserved(self):
        """Placeholders correctly preserved"""
        result = await self.classify(
            "Amount received: {}",
            "Montant reçu : {}",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Placeholder preservation should be Genuine")

    async def test_genuine_qt_markup_preserved(self):
        """Qt markup preserved in translation"""
        result = await self.classify(
            "&File",
            "&Fichier",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Qt markup preservation should be Genuine")

    async def test_genuine_html_markup(self):
        """HTML markup in translation"""
        result = await self.classify(
            "<b>Warning:</b> This action cannot be undone",
            "<b>Avertissement :</b> Cette action ne peut pas être annulée",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "HTML markup translation should be Genuine")

    async def test_genuine_technical_domain_terms(self):
        """Domain-specific technical terms"""
        result = await self.classify(
            "CPFP transaction",
            "Transaction CPFP",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Technical domain terms should be Genuine")

    async def test_genuine_bitcoin_terminology(self):
        """Bitcoin-specific terminology preserved"""
        result = await self.classify(
            "Satoshis per byte",
            "Satoshis par octet",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Bitcoin terminology should be Genuine")

    async def test_genuine_long_instruction_text(self):
        """Longer instructional text"""
        result = await self.classify(
            "A CPFP is a transaction that sends an unconfirmed output back to yourself, with a high fee.",
            "Un CPFP est une transaction qui vous renvoie une sortie non confirmée, avec des frais élevés.",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Long translated text should be Genuine")

    async def test_genuine_idiomatic_translation(self):
        """Idiomatic translation that differs from literal"""
        result = await self.classify(
            "Got it",
            "D'accord",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Idiomatic translations should be Genuine")


class TestEdgeCases(AsyncVandalismTestCase):
    """Tests for edge cases and boundary conditions."""

    async def test_partial_translation_genuine(self):
        """Partial translation should still be Genuine"""
        result = await self.classify(
            "Bitcoin wallet backup",
            "Bitcoin wallet サックアップ",
            "ja_JP"
        )
        self.assertEqual(result, "Genuine", "Partial translation should be Genuine")

    async def test_abbreviation_preserved(self):
        """Common abbreviation preserved"""
        result = await self.classify(
            "BTC",
            "BTC",
            "es_ES"
        )
        self.assertEqual(result, "Genuine", "Preserved abbreviation should be Genuine")

    async def test_number_format_different(self):
        """Number formatting can vary by locale"""
        result = await self.classify(
            "1,000.00 BTC",
            "1.000,00 BTC",
            "de_DE"
        )
        self.assertEqual(result, "Genuine", "Locale-specific number format should be Genuine")


class TestVandalismSeedPhraseInjection(AsyncVandalismTestCase):
    """Tests for seed phrase / private key injection attempts."""

    async def test_seed_phrase_injection(self):
        """Seed phrase words injected as translation"""
        result = await self.classify(
            "Enter your password",
            "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "Seed phrase injection should be Spam")

    async def test_private_key_wif_injection(self):
        """WIF private key injected"""
        result = await self.classify(
            "Backup wallet",
            "5HueCGU8rMjxEXxiPuD5BDku4MkFqeZyd4dZ1jvhTVqvbTLvyTJ",
            "fr_FR"
        )
        self.assertEqual(result, "Spam", "Private key injection should be Spam")

    async def test_hex_private_key_injection(self):
        """Hex private key injected"""
        result = await self.classify(
            "Sign message",
            "e8f32e723decf4051aefac8e2c93c9c5b214313817cdb01a1494b917c8436b35",
            "it_IT"
        )
        self.assertEqual(result, "Spam", "Hex key injection should be Spam")


class TestVandalismOffensiveContent(AsyncVandalismTestCase):
    """Tests for offensive / inappropriate content."""

    async def test_profanity_injection(self):
        """Profanity injected instead of translation"""
        result = await self.classify(
            "Loading wallet...",
            "Fuck this stupid app",
            "en_GB"
        )
        self.assertEqual(result, "Spam", "Profanity injection should be Spam")

    async def test_insult_injection(self):
        """Personal insult injected"""
        result = await self.classify(
            "Transaction complete",
            "You are an idiot for using this",
            "es_ES"
        )
        self.assertEqual(result, "Spam", "Insult injection should be Spam")


class TestVandalismPromotionalContent(AsyncVandalismTestCase):
    """Tests for promotional / marketing spam."""

    async def test_casino_promotion(self):
        """Online casino promotion injected"""
        result = await self.classify(
            "Fee estimation",
            "Win big at CryptoSlots Casino! Deposit now!",
            "pt_BR"
        )
        self.assertEqual(result, "Spam", "Casino promotion should be Spam")

    async def test_investment_scam(self):
        """Investment scam text injected"""
        result = await self.classify(
            "Verify signature",
            "Double your Bitcoin in 24 hours! Send to this address!",
            "ru_RU"
        )
        self.assertEqual(result, "Spam", "Investment scam should be Spam")

    async def test_job_offer_spam(self):
        """Fake job offer spam"""
        result = await self.classify(
            "Address type",
            "Work from home $5000/week contact john@gmail.com",
            "nl_NL"
        )
        self.assertEqual(result, "Spam", "Job offer spam should be Spam")


class TestVandalismIdentityInjection(AsyncVandalismTestCase):
    """Tests for personal identity / name injection."""

    async def test_random_person_name(self):
        """Random person name injected"""
        result = await self.classify(
            "&Copy",
            "John Smith",
            "zh_CN"
        )
        self.assertEqual(result, "Spam", "Random name injection should be Spam")

    async def test_company_name_injection(self):
        """Unrelated company name injected"""
        result = await self.classify(
            "Block height",
            "Microsoft Corporation",
            "ja_JP"
        )
        self.assertEqual(result, "Spam", "Company name injection should be Spam")

    async def test_gamer_tag_injection(self):
        """Gaming handle / gamer tag injected"""
        result = await self.classify(
            "Sweep all",
            "xXx_Pr0Gamer_420_xXx",
            "ko_KR"
        )
        self.assertEqual(result, "Spam", "Gamer tag injection should be Spam")


class TestVandalismDateTimeInjection(AsyncVandalismTestCase):
    """Tests for arbitrary date/time injection."""

    async def test_random_date_injection(self):
        """Random date injected"""
        result = await self.classify(
            "Confirm payment",
            "2025-03-15 14:32:00",
            "sv_SE"
        )
        self.assertEqual(result, "Spam", "Random date injection should be Spam")

    async def test_countdown_injection(self):
        """Countdown text injected"""
        result = await self.classify(
            "Sign transaction",
            "Only 3 days left until the big event!",
            "da_DK"
        )
        self.assertEqual(result, "Spam", "Countdown injection should be Spam")


class TestVandalismCodeInjection(AsyncVandalismTestCase):
    """Tests for code/script injection attempts."""

    async def test_javascript_injection(self):
        """JavaScript code injected"""
        result = await self.classify(
            "Click to continue",
            "<script>alert('Klicken um fortzufahren')</script>",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "JavaScript injection should be Spam")

    async def test_qt_markup_injection(self):
        """Qt markup injected"""
        result = await self.classify(
            "Click to continue",
            "<a href='https://electrum.io/landing'>Klicken um fortzufahren</a>",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "Qt markup injection should be Spam")

    async def test_sql_injection_attempt(self):
        """SQL injection attempt"""
        result = await self.classify(
            "Search",
            "'; DROP TABLE wallets; --",
            "pl_PL"
        )
        self.assertEqual(result, "Spam", "SQL injection should be Spam")


class TestGenuineComplexTranslations(AsyncVandalismTestCase):
    """Tests for complex genuine translations that might be edge cases."""

    async def test_genuine_with_numbers(self):
        """Translation containing legitimate numbers"""
        result = await self.classify(
            "Requires 2 confirmations",
            "Requiert 2 confirmations",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Translation with numbers should be Genuine")

    async def test_genuine_plural_form(self):
        """Plural form translation"""
        result = await self.classify(
            "{} confirmations",
            "{} confirmations",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Plural form should be Genuine")

    async def test_genuine_empty_placeholder(self):
        """String that is just a placeholder"""
        result = await self.classify(
            "%s",
            "%s",
            "de_DE"
        )
        self.assertEqual(result, "Genuine", "Placeholder-only string should be Genuine")

    async def test_genuine_mixed_scripts_thai(self):
        """Thai translation with English technical terms"""
        result = await self.classify(
            "Bitcoin wallet",
            "กระเป๋า Bitcoin",
            "th_TH"
        )
        self.assertEqual(result, "Genuine", "Thai translation should be Genuine")

    async def test_genuine_mixed_scripts_hebrew(self):
        """Hebrew translation with mixed direction"""
        result = await self.classify(
            "Send BTC",
            "שלח BTC",
            "he_IL"
        )
        self.assertEqual(result, "Genuine", "Hebrew translation should be Genuine")

    async def test_genuine_mixed_scripts_hindi(self):
        """Hindi translation"""
        result = await self.classify(
            "Wallet",
            "वॉलेट",
            "hi_IN"
        )
        self.assertEqual(result, "Genuine", "Hindi translation should be Genuine")

    async def test_genuine_transliteration(self):
        """Transliteration is acceptable"""
        result = await self.classify(
            "Bitcoin",
            "比特币",
            "zh_CN"
        )
        self.assertEqual(result, "Genuine", "Transliteration should be Genuine")

    async def test_genuine_complex_qt_markup(self):
        """Complex Qt markup with multiple elements"""
        result = await self.classify(
            "&Edit | &Delete | &Copy",
            "&Éditer | &Supprimer | &Copier",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Complex Qt markup should be Genuine")

    async def test_genuine_multiple_placeholders(self):
        """Multiple placeholders in string"""
        result = await self.classify(
            "Sent {amount} to {address}",
            "Envoyé {amount} à {address}",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Multiple placeholders should be Genuine")

    async def test_genuine_error_message(self):
        """Error message translation"""
        result = await self.classify(
            "Error: Insufficient funds",
            "Erreur : Fonds insuffisants",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Error message translation should be Genuine")

    async def test_genuine_question_text(self):
        """Question text translation"""
        result = await self.classify(
            "Are you sure you want to delete this wallet?",
            "Êtes-vous sûr de vouloir supprimer ce portefeuille ?",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Question translation should be Genuine")

    async def test_genuine_url_in_original(self):
        """URL present in original should be preserved"""
        result = await self.classify(
            "Visit https://electrum.org for more information",
            "Visitez https://electrum.org pour plus d'informations",
            "fr_FR"
        )
        self.assertEqual(result, "Genuine", "Preserved URL should be Genuine")


class TestVandalismMoreCryptoScams(AsyncVandalismTestCase):
    """Additional tests for crypto-specific scams."""

    async def test_fake_support_message(self):
        """Fake support message injection"""
        result = await self.classify(
            "Help",
            "For support contact admin@electrum-support.com",
            "de_DE"
        )
        self.assertEqual(result, "Spam", "Fake support message should be Spam")

    async def test_airdrop_scam(self):
        """Airdrop scam injection"""
        result = await self.classify(
            "Receive address",
            "Claim your free 1 BTC airdrop now!",
            "es_ES"
        )
        self.assertEqual(result, "Spam", "Airdrop scam should be Spam")

    async def test_exchange_promotion(self):
        """Exchange promotion injection"""
        result = await self.classify(
            "Exchange rate",
            "Trade on BinanceScam exchange with 0% fees!",
            "zh_TW"
        )
        self.assertEqual(result, "Spam", "Exchange promotion should be Spam")


class TestVandalismSpecialCharacters(AsyncVandalismTestCase):
    """Tests for special character abuse."""

    async def test_zalgo_text(self):
        """Zalgo/corrupted text injection"""
        result = await self.classify(
            "Balance",
            "B̷̢̺̈́ą̸̛l̷̰͑a̸̱͝n̵̗̏c̸̣̊e̷̤̔",
            "en_US"
        )
        self.assertEqual(result, "Spam", "Zalgo text should be Spam")

    async def test_invisible_characters(self):
        """Zero-width characters abuse"""
        result = await self.classify(
            "Send",
            "S\u200be\u200bn\u200bd\u200b\u200b\u200bClick here for free BTC",
            "fr_FR"
        )
        self.assertEqual(result, "Spam", "Hidden text injection should be Spam")

    async def test_excessive_punctuation(self):
        """Excessive punctuation spam"""
        result = await self.classify(
            "Warning!",
            "!!!!!!!!!!!!!!!!!!!!",
            "it_IT"
        )
        self.assertEqual(result, "Spam", "Excessive punctuation should be Spam")


class TestVandalismIPAddresses(AsyncVandalismTestCase):
    """Tests for IP address injection."""

    async def test_ipv4_injection(self):
        """IPv4 address injected"""
        result = await self.classify(
            "Server settings",
            "192.168.1.100:8080",
            "ru_RU"
        )
        self.assertEqual(result, "Spam", "IPv4 injection should be Spam")

    async def test_ipv6_injection(self):
        """IPv6 address injected"""
        result = await self.classify(
            "Connect to node",
            "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
            "ja_JP"
        )
        self.assertEqual(result, "Spam", "IPv6 injection should be Spam")


if __name__ == "__main__":
    unittest.main(verbosity=2)




#!/usr/bin/env python3
import sys
import asyncio

from electrum.network import filter_protocol, Network
from electrum.util import create_and_start_event_loop, log_exceptions
from electrum.simple_config import SimpleConfig


try:
    txid = sys.argv[1]
except Exception:
    print("usage: txradar txid")
    sys.exit(1)

config = SimpleConfig()

loop, stopping_fut, loop_thread = create_and_start_event_loop()
network = Network(config)
network.start()

@log_exceptions
async def f():
    try:
        peers = await network.get_peers()
        peers = filter_protocol(peers)
        results = await network.send_multiple_requests(peers, 'blockchain.transaction.get', [txid])
        r1, r2 = [], []
        for k, v in results.items():
            (r1 if not isinstance(v, Exception) else r2).append(k)
        print(f"Received {len(results)} answers")
        try: propagation = len(r1) * 100. / (len(r1) + len(r2))
        except ZeroDivisionError: propagation = 0
        print(f"Propagation rate: {propagation:.1f} percent")
    finally:
        stopping_fut.set_result(1)

asyncio.run_coroutine_threadsafe(f(), loop)
while loop_thread.is_alive():
    loop_thread.join(1)



from typing import Optional

import qrcode
import qrcode.exceptions

import PyQt6.QtGui as QtGui
from PyQt6.QtCore import QRect
from PyQt6.QtWidgets import QApplication, QVBoxLayout, QHBoxLayout, QPushButton, QWidget

from electrum.i18n import _
from electrum.simple_config import SimpleConfig
from electrum.gui.common_qt.util import draw_qr

from .util import WindowModalDialog, WWLabel, getSaveFileName


class QrCodeDataOverflow(qrcode.exceptions.DataOverflowError):
    pass


class QRCodeWidget(QWidget):

    MIN_BOXSIZE = 2  # min size in pixels of single black/white unit box of the qr code

    def __init__(self, data=None, *, manual_size: bool = False):
        QWidget.__init__(self)
        self.data = None
        self.qr = None
        self._framesize = None  # type: Optional[int]
        self._manual_size = manual_size
        self.setData(data)

    def setData(self, data):
        if data:
            qr = qrcode.QRCode(
                error_correction=qrcode.constants.ERROR_CORRECT_L,
                border=1,
            )
            try:
                qr.add_data(data)
                qr_matrix = qr.get_matrix()  # test that data fits in QR code
            except (ValueError, qrcode.exceptions.DataOverflowError) as e:
                raise QrCodeDataOverflow() from e
            self.qr = qr
            self.data = data
            if not self._manual_size:
                k = len(qr_matrix)
                size = min(k * 5, 150 + k * self.MIN_BOXSIZE)
                self.setMinimumSize(size, size)
        else:
            self.qr = None
            self.data = None

        self.update()

    def paintEvent(self, e):
        if not self.data:
            return
        draw_qr(
            qr=self.qr,
            paint_device=self,
            is_enabled=self.isEnabled(),
            min_boxsize=self.MIN_BOXSIZE,
        )

    def grab(self) -> QtGui.QPixmap:
        """Overrides QWidget.grab to only include the QR code itself,
        excluding horizontal/vertical stretch.
        """
        fsize = self._framesize
        if fsize is None:
            fsize = -1
        rect = QRect(0, 0, fsize, fsize)
        return QWidget.grab(self, rect)


class QRDialog(WindowModalDialog):

    def __init__(
            self,
            *,
            data,
            parent=None,
            title="",
            show_text=False,
            help_text=None,
            show_copy_text_btn=False,
            config: SimpleConfig,
    ):
        WindowModalDialog.__init__(self, parent, title)
        self.config = config

        vbox = QVBoxLayout()

        qrw = QRCodeWidget(data, manual_size=False)
        vbox.addWidget(qrw, 1)

        help_text = data if show_text else help_text
        if help_text:
            text_label = WWLabel()
            text_label.setText(help_text)
            vbox.addWidget(text_label)
        hbox = QHBoxLayout()
        hbox.addStretch(1)

        def print_qr():
            filename = getSaveFileName(
                parent=self,
                title=_("Select where to save file"),
                filename="qrcode.png",
                config=self.config,
            )
            if not filename:
                return
            p = qrw.grab()
            p.save(filename, 'png')
            self.show_message(_("QR code saved to file") + " " + filename)

        def copy_image_to_clipboard():
            p = qrw.grab()
            QApplication.clipboard().setPixmap(p)
            self.show_message(_("QR code copied to clipboard"))

        def copy_text_to_clipboard():
            QApplication.clipboard().setText(data)
            self.show_message(_("Text copied to clipboard"))

        b = QPushButton(_("Copy Image"))
        hbox.addWidget(b)
        b.clicked.connect(copy_image_to_clipboard)

        if show_copy_text_btn:
            b = QPushButton(_("Copy Text"))
            hbox.addWidget(b)
            b.clicked.connect(copy_text_to_clipboard)

        b = QPushButton(_("Save"))
        hbox.addWidget(b)
        b.clicked.connect(print_qr)

        b = QPushButton(_("Close"))
        hbox.addWidget(b)
        b.clicked.connect(self.accept)
        b.setDefault(True)

        vbox.addLayout(hbox)
        self.setLayout(vbox)

        # note: the word-wrap on the text_label is causing layout sizing issues.
        #       see https://stackoverflow.com/a/25661985 and https://bugreports.qt.io/browse/QTBUG-37673
        #       workaround:
        self.setMinimumSize(self.sizeHint())
